#ifndef ___HGL___smart_ptr___
#define ___HGL___smart_ptr___

#include "ref_delete.h"
#include "function_callback.h"

/**
 smart_ptr_base
 スマートポインタの基底クラス。
 特に何をするわけでもない。
*/
class smart_ptr_base {
public:
	virtual ~smart_ptr_base(){}
};

/**
 smart_ptr
 スマートポインタ。
 smart_ptr<CHoge> h(new CHoge);
 とやれば、後はコピーしても解体責任を委譲したりして、
 ちゃんとdeleteしてくれるようなクラス。
 また、配列もとれるようにして、配列であっても
 通常のポインタであっても、ちゃんとdeleteとdelete[]を
 分けてくれる。

 配列を生成する際には、

 smart_ptr<CHoge> h;
 h.addArray(true,10);
 のように生成するか、
 もしくは、CHogeの派生クラスSHogeを生成したりする場合は、
 h.addArray(new SHoge[10],true,10);

 のようにする。
*/
template<class T>
class smart_ptr : public smart_ptr_base {
public:

	//暗黙の型変換を行わない。
	//このコンストラクタでは、配列の生成は出来ないことになっている。
	template<class S> explicit 
		smart_ptr(S* p, bool owner = true, ref_delete_base* ref = NULL) : 
			m_pObj(NULL),m_pRefObj(NULL),m_nIndex(0) {
		if (p != NULL) {
			init(p,owner,1, ref);
		}
	}

	//デフォルトコンストラクタ
	smart_ptr() : m_pObj(NULL),m_pRefObj(NULL),m_nIndex(0) {
	}
	
	//コピーコンストラクタ
	smart_ptr(const smart_ptr<T>& p) : 
	m_pObj(NULL),m_pRefObj(NULL),m_nIndex(0) {
		//参照カウントを増やしつつ、コピー。
		inc(p);
	}

	//参照数を減らすだけ。
	virtual ~smart_ptr() {dec();}

	//明示的に削除するための構文
	virtual void release() {dec();}

	//ここからは、ポインタの振りをするための仕掛け。
	T& operator*() const {return *get();}
	T* operator->() const {return get();}

	//擬似的な配列操作のため。
	smart_ptr<T>& operator++() {incIndex();return *this;}
	smart_ptr<T>& operator++(int n) {incIndex();return *this;}
	smart_ptr<T>& operator--() {decIndex();return *this;}
	smart_ptr<T>& operator--(int n) {decIndex();return *this;}

	//配列からの取得のため。
	T& operator[](int n) const {return *get(n + 1);}

	smart_ptr<T>& operator=(const smart_ptr<T>& obj) {
		//コピーコンストラクタ。
		//前のオブジェクトの参照カウントを減らした後、
		//コピーする
		if (this != &obj) {
			dec();
			inc(obj);
		}
		return *this;
	}

	/**
	 get
	 保持するメモリから、指定された場所のメモリを確保する。
	 @param int n デフォルトは1
	 @access public
	 @return 指定された位置のテンプレートで指定されたオブジェクト
	*/
	T* get(int n = 1) const {
		//配列、単一変数どちらにでも対応するため、
		//引数で場所を取る。
		

		if (m_pRefObj == NULL) {
			return NULL;
		}

		if (m_nIndex + n > m_pRefObj->getMaxNum() || m_nIndex + n < 0) {
			return NULL;
		}

		//取得すべきメモリは、
		//オブジェクトの先頭アドレス+オブジェクトのサイズ*(インデックス+引数)
		//なお、この際に、m_nIndex = 0かつ、n = 1である場合、
		//先頭アドレス+オブジェクトサイズということになり、どう考えてもアクセス違反になる。
		//である
		return (T*)((unsigned char*)m_pObj + (m_pRefObj->getObjSize() * (m_nIndex + n - 1)));
	}

	//==============================================
	//疑似イテレーターとして使用するための関数群
	//==============================================

	/**
	 isEnd
	 インデックスが、限界に達していないかどうか
	 ここで言う限界とは、MaxNumで、つまり添え字+1をさす。
	 @access public
	 @return bool 達している場合はtrue
	*/
	bool isEnd() {
		if (m_pRefObj != NULL) {
			return m_pRefObj->getMaxNum()-1 < m_nIndex?true:false;
		} else {
			return true;
		}
	}

	/**
	 isBegin
	 インデックスが、先頭であるかどうか
	 @access public
	 @return bool 先頭である場合はtrue
	*/
	bool isBegin() {
		return m_nIndex == 0?true:false;
	}

	/**
	 オブジェクトがNULLかどうか
	 @access public
	 @return bool
	*/
	bool isNull() const {
		return m_pRefObj==NULL&&m_pObj==NULL?true:false;
	}
	
	/**
	 end
	 インデックスを、最後尾に位置させる
	 @access public
	 @return void
	*/
	void end() {
		if (m_pRefObj != NULL) {
			m_nIndex = m_pRefObj->getMaxNum()-1;
		} else {
			m_nIndex = 0;
		}
	}

	/**
	 begin
	 インデックスを、先頭に位置させる
	 @access public
	 @return void
	*/
	void begin() {
		m_nIndex = 0;
	}

	//=============================================
	//追加生成関数群
	//=============================================

	/**
	 * @brief アップキャストを行う。SはTの派生クラスであること。
	 * また、その特性上同一性チェックが困難であるので、こればかりは仕方がない。
	 * @param const smart_ptr<S>& スマートポインタ
	 * @return smart_ptr<T>& アップキャスト後の自身。
	 */
	template<class S>
	smart_ptr<T>& upcast(const smart_ptr<S>& p) {
		
		if (reinterpret_cast<void*>(this) != 
			reinterpret_cast<void*>(const_cast<smart_ptr<S>*>(&p)) ) {
			//pがさすオブジェクトをこのオブジェクトもさすことになるので、
			//まずはdec
			dec();
			
			//あとはincとほぼ同じ処理になる。
			m_pObj = p.getObj();
			m_pRefObj = p.getRefObj();
			m_nIndex = p.getIndex();

			if (m_pRefObj != NULL) {
				m_pRefObj->inc_ref();
			}
		}

		return *this;
	}

	/**
	 * @brief 指定した型の追加生成構文
	 * コンストラクタで何も指定しない場合、これを実行することで、追加生成か可能となる。
	 */
	void add() {
		dec();
		init(new T, true, 1);
	}

	/**
	 add
	 新しく生成させる。
	 関数テンプレートを使ってはいるが、必ずTとおなじ、
	 もしくはTの派生クラスであること。
	 @param S* p 対象とするオブジェクト
	 @param bool owner 所有権の有無
	 @access public
	 @return void
	*/
	template<class S>
	void add(S* _p, bool owner = true) {
		dec();
		if (_p != NULL) {
			init(_p,owner,1);
		}
	}

	/**
	 * @brief デリータを指定した追加生成構文。
	 * これは配列では使用できない。
	 */
	template<class S>
	void add(S* _p, bool owner, ref_delete_base* ref) {
		dec();
		if (_p != NULL && ref != NULL) {
			init(_p, owner, 1, ref);
		}
	}

	/**
	 addArray
	 配列オブジェクト生成関数
	 現在のオブジェクトの型を、指定個数作成する場合。
	 @param bool owner 所有権の有無
	 @param int n 個数
	 @access public
	 @return void
	*/
	void addArray(bool owner, int n) { 
		dec();
		if (n > 1) {
			init(new T[n], owner, n);
		} else if (n == 1) {
			init(new T, owner, 1);
		}
	}

	/**
	 addArray
	 配列オブジェクト生成関数
	 こちらは、配列オブジェクトを生成する際に使用される関数。
	 @param S p 対象とするオブジェクト
	 @param bool owner 所有権の有無
	 @param int n 個数
	 @access public
	 @return void
	*/
	template<class S>
	void addArray(S _p, bool owner, int n) { 
		dec();
		if (n > 1) {
			init(new S[n], owner, n);
		}
	}

	/**
	 getIndex
	 現在のインデックスを返す
	 @access public
	 @return int 現在のインデックス
	*/
	int getIndex() const {return m_nIndex;}

	/**
	 getRefObj
	 削除オブジェクトを返す
	 @access public
	 @return ref_delete_base* 削除オブジェクトのポインタ
	*/
	ref_delete_base* getRefObj() const {return m_pRefObj;}
	
	/**
	 getObj
	 オブジェクトを返す
	 @access public
	 @return T* オブジェクトのポインタ
	*/
	T* getObj() const {return m_pObj;}

private:

	/**
	 init
	 スマートポインタの初期化補助関数
	 渡されたオブジェクトを元にして、所有権、
	 参照カウント、オブジェクトの個数を設定する。
	 @param template _p 保持するべきオブジェクト
	 @param bool owner 所有権の湯有無
	 @param int n オブジェクトの個数
	 @param ref_delete_base* ref デフォルトはNULL。外部からのコピーならば非NULL
	 @access protected
	 @return void
	*/
	template<class S>
	void init(S* _p, bool owner, int n, ref_delete_base* ref = NULL) {
		
		m_nIndex = 0;
		m_pObj = _p;

		if (m_pRefObj != NULL) {
			//このとき、すでにdec_refとかはやってあることが前提。
			m_pRefObj = NULL;
		}

		if (ref != NULL) {
			m_pRefObj = ref;
		} else if (owner) {
			if (n == 1) {
				//非配列オブジェクト
				m_pRefObj = new ref_noarray_object<S>(_p);
			} else if (n > 1){
				m_pRefObj = new ref_array_object<S>(_p);
			}
		} else {
			//非所有権オブジェクト
			m_pRefObj = new ref_null_object<S>(_p);
		}

		//オブジェクトの情報を追加
		m_pRefObj->setRef(1);
		m_pRefObj->setMaxNum(n);
		m_pRefObj->setObjSize(sizeof(S));
		m_pRefObj->setOwner(owner);

	}

	/**
	 inc
	 渡されたsmart_ptrの参照数を一つ増やす。
	 @param const smart_ptr<T>& p 参照カウントを増やすオブジェクト
	 @access protected
	 @return void
	*/
	void inc(const smart_ptr<T>& p) {
		m_nIndex	= p.getIndex();
		m_pRefObj	= p.getRefObj();
		m_pObj		= p.getObj();
		
		if (m_pRefObj != NULL) {
			m_pRefObj->inc_ref();
		}
	}

	/**
	 dec
	 自分の分の参照カウントを減らす。
	 実行されたら、m_pRefObjおよびm_pObjは
	 NULLになる。
	 @access protected
	 @return void
	*/
	void dec() {
		//dec_refは、オブジェクトを削除したかどうかを返す。
		//削除されている場合は、参照オブジェクトを削除しなければならない。
		if (m_pRefObj != NULL) {
			if (m_pRefObj->dec_ref()) {
				delete m_pRefObj;
			}
		}
		m_pRefObj = NULL;
		m_pObj = NULL;
	}

	/**
	 incIndex
	 インデックスカウントを増やす
	 @access protected
	 @return void
	*/
	void incIndex() {m_nIndex++;}

	/**
	 decIndex
	 インデックスカウントを減らす
	 @access protected
	 @return void
	*/
	void decIndex() {m_nIndex--;}

	//========================================
	//コピーの際に使用するconst関数群
	//========================================
	
private:

	T* m_pObj;		//保持すべきオブジェクト。参照カウントオブジェクトと同一

	ref_delete_base* m_pRefObj;		//参照カウントオブジェクト

	int m_nIndex;		//現在のインデックス

};

template<class T>
class ref_noarray_func_object : public ref_delete_base {
public:

	ref_noarray_func_object(T* p, const smart_ptr<function_callback>& func) :
		m_pObj(p), m_pFunc(func)
	{
	}

	virtual void do_delete() {
		m_pFunc->run();
		if (m_pObj != NULL) {
			delete m_pObj;
		}
	}

private:

	T* m_pObj;
	smart_ptr<function_callback> m_pFunc;
};
				
#endif