#include <sstream>
#include <iomanip>

#include <algorithm>
#include <utility>
#include <vector>
#include <cctype>

#include "bignum.h"

#include "debugutil.h"
namespace debug = utakata::debugutil;

namespace nume = utakata::numeric;

struct nume::Bignum::pImpl
{
  pImpl() : flag(1) {}
  std::vector<int> digits;
  int flag;                   // フラグを1と-1で表現する。
};

nume::Bignum::Bignum() : num_(new nume::Bignum::pImpl)
{
  // 初期の場合、digitsには0のみが設定される。
  num_->digits.push_back(0);
}

nume::Bignum::Bignum(const nume::Bignum::Bignum& rh) : num_(new nume::Bignum::pImpl)
{
  std::vector<int> p = rh.num_->digits;
  std::swap(p, num_->digits);
  num_->flag = rh.num_->flag;
}

nume::Bignum::Bignum(const std::string& str) : num_(new nume::Bignum::pImpl)
{
  // 文字列を簡単に解析し、文字列として保存する。
  parse_string_(str);
}

nume::Bignum& nume::Bignum::operator=(const nume::Bignum& rh)
{
  ////////////////////////////
  // 代入は簡単に行われる。 //
  ////////////////////////////

  nume::Bignum tmp(rh);
  swap_(tmp);
  return *this;
}

bool nume::Bignum::operator<(const nume::Bignum& rh) const
{
  //////////////////////////////////////////////////////////////////////////////////
  // 引数で渡されたBignumよりも自身が小さい場合にtrueを返す。                     //
  // trueを返す条件としては次のいずれかの条件を満たした場合となる。               //
  // ・符号が小さい(rh側が正である場合のみ)                                       //
  // ・桁が小さい(正の場合) または 桁が大きい(負の場合)                           //
  // ・数値の先頭から順に検査(正の場合、小さいかどうか、負の場合、大きいかどうか) //
  //////////////////////////////////////////////////////////////////////////////////

  if (num_->flag < rh.num_->flag)
  {
    return true;
  }
  else if (num_->flag > rh.num_->flag)
  {
    return false;
  }

  if (less_between<int>(num_->flag)(num_->digits.size(),
                                    rh.num_->digits.size()))
  {
    return true;
  }
  else if (greater_between<int>(num_->flag)(num_->digits.size(),
                                            rh.num_->digits.size()))
  {
    return false;
  }

  std::vector<int>& t = num_->digits;
  std::pair<std::vector<int>::reverse_iterator,
    std::vector<int>::reverse_iterator> p =
    std::mismatch(t.rbegin(),
                  t.rend(),
                  rh.num_->digits.rbegin(),
                  std::equal_to<int>());
  // predicateに適合しなかったデータがあり、それが小さい場合にのみ、
  // 該当する。
  // このoperatorに適合しない。
  if (p.first != t.rend() && less_between<int>(num_->flag)(
          *p.first, *p.second))
  {
    return true;
  }
  return false;
}

bool nume::Bignum::operator==(const nume::Bignum& rh) const
{
  //////////////////////////////////////////////////////////
  // 符号、桁数、各桁が等しい場合にのみ、等しいとされる。 //
  //////////////////////////////////////////////////////////
  if (num_->flag != rh.num_->flag)
  {
    return false;
  }
  else if (num_->digits.size() != rh.num_->digits.size())
  {
    return false;
  }

  std::vector<int>& t = num_->digits;
  it_pair p = std::mismatch(
      t.begin(),
      t.end(),
      rh.num_->digits.begin(),
      std::equal_to<int>());

  if (p.first == t.end())
  {
    return true;
  }

  return false;
}

std::string nume::Bignum::to_str() const
{
  ////////////////////////////////////////////////////////////////////////////////////
  // 内部の文字列を数値として利用する。内部の数値は、次の手順で文字列に変換される。 //
  // 1. 符号が-1である場合、-を先頭に付与する                                       //
  // 2. digitsの末尾を、「0詰めなし」で文字列化し、出力する。                       //
  // 3. digitsの末尾以降を、「0詰めあり」で文字列化する。                           //
  ////////////////////////////////////////////////////////////////////////////////////

  std::stringstream ret;
    
  if (num_->flag < 0)
  {
    ret << "-";
  }

  std::vector<int>::reverse_iterator b = num_->digits.rbegin(),
    e = num_->digits.rend();
  ret << *b;
  ++b;

  std::for_each(b, e, nume::Bignum::convert_str(digit_unit_size_, ret));

  return ret.str();
}

void nume::Bignum::normalize_()
{
  ///////////////////////////////////////////////////////////////////////////////
  // 現在保持しているdigitsを、桁毎に最高値(digit_unit)を下回るように          //
  // 補正する。                                                                //
  // 補正は、以下のようなアルゴリズムで行われる。                              //
  // 1.データの絶対値がdigit_unitを超えている場合、digit_unit単位での          //
  //   調整を行う。調整はデータの符号単位で行う。                              //
  // 2.現在の処理中の桁に対して、次の桁の符号が反転している場合、              //
  //   次の桁に、次の桁の符号を反転させた1を加算し、現在の桁に、現在の桁の符号 //
  //   を反転させたdigit_unitを加算する。                                      //
  // いずれの場合においても、現在処理中の桁に限り、補正後に絶対値化を行う。    //
  ///////////////////////////////////////////////////////////////////////////////

  std::vector<int>& tmp = num_->digits;

  std::vector<int>::iterator it = tmp.begin(), fw, 
    end = tmp.begin();
  std::advance(end, tmp.size() - 1);

  for (; it != end; ++it)
  {
    fw = it + 1;
    normalize_one_(it, it + 1);
  }

  // 先頭については言及されていないため、ここ言及する。
  int nowsize = tmp.size();
  while (std::abs(*end) >= digit_unit_)
  {
    if (end + 1 == tmp.end())
    {
      tmp.push_back(to_flag_(*end));
      end = tmp.begin(); std::advance(end, nowsize + 1);
    }

    normalize_one_(end, end + 1);
    ++end;
  }

  {
    // 先頭から続く0を削除。
    std::vector<int> rev = tmp;
    std::reverse(rev.begin(), rev.end());
    std::vector<int>::iterator fw =
      std::find_if(rev.begin(), rev.end(),
                   equal_zero());
    if (rev.begin() != fw)
    {
      rev.erase(rev.begin(), fw);
    }

    std::reverse(rev.begin(), rev.end());
    tmp = rev;
  }

  end = tmp.begin(); std::advance(end, tmp.size() - 1);
  num_->flag = to_flag_(*end);
  *end = std::abs(*end);
}

void nume::Bignum::normalize_one_(std::vector<int>::iterator it,
                                  std::vector<int>::iterator fwd)
{
  //////////////////////////////////////////////////////////////////
  // 単一の桁に対する正規化処理を行う。この処理はイテレータ抽象で //
  // 完全に処理が可能であるため、単一の関数に分割している。       //
  //////////////////////////////////////////////////////////////////
  while (std::abs(*it) >= digit_unit_)
  {
    int rflag = reverse_flag_(*it), flag = to_flag_(*it);
    *it += rflag * digit_unit_;
    *fwd += flag;
  }

  if (to_flag_(*fwd) != to_flag_(*it) && *it != 0)
  {
    int f1 = reverse_flag_(*fwd), f2 = reverse_flag_(*it);
    *it += f1 * digit_unit_;
    *fwd += f2;
  }
  *it = std::abs(*it);
}

void nume::Bignum::swap_(Bignum& rh)
{
  ///////////////////////////////////////////////////////////////////////////////
  // 内容を交換する。内容として保持しているのは、pImplのみであるため、そのまま //
  // 利用することができる。                                                    //
  ///////////////////////////////////////////////////////////////////////////////
  smart_ptr<pImpl> p = rh.num_;
  rh.num_ = num_;
  num_ = p;
}

void nume::Bignum::parse_string_(const std::string& str)
{
  std::string t;
  if (*str.begin() == '-')
  {
    num_->flag = -1;
    t.append(str.begin() + 1, str.end());
  }
  else
  {
    num_->flag = 1;
    t = str;
  }

  std::string::iterator type =
    std::find_if(t.begin(), t.end(),
                 nume::Bignum::not_equal_digits());

  if (type != t.end())
  {
    throw "Don't charactor that not digits is contain to parse string!";
  }

  // 先頭から随時データを抽出する。
  std::vector<int> swaping;
  for (int i = t.size() ; i > 0; i -= digit_unit_size_)
  {
    if (i < digit_unit_size_)
    {
      swaping.push_back(
          atoi(t.substr(0, i).c_str()));
    }
    else
    {
      swaping.push_back(
          atoi(t.substr(i - digit_unit_size_, digit_unit_size_).c_str()));
    }
  }

  std::swap(num_->digits, swaping);
}

nume::Bignum& nume::Bignum::operator+=(const nume::Bignum& rh)
{
  /////////////////////////////////////////////////////
  // 加算は、次のアルゴリズムによって行われる。      //
  // 1. rh, lhのそれぞれの各digitsに符号を乗算する   //
  // 2. lh, rhの桁数を合わせる。増えた分は0とする    //
  // 3. lh, rhをそれぞれ加算し、結果をlhに設定する。 //
  // 4. lhのnormalizeを行う。                        //
  /////////////////////////////////////////////////////

  nume::Bignum tmprh = rh;
  std::vector<int>& tmpt = num_->digits;
  std::vector<int>& tmpr = tmprh.num_->digits;
  {
    size_t m = std::max(num_->digits.size(),
                        rh.num_->digits.size());
    tmpr.resize(m);
    tmpt.resize(m);
  }

  std::transform(tmpr.begin(), tmpr.end(),
                 tmpr.begin(),
                 nume::Bignum::modulu_unary(rh.num_->flag));
  std::transform(tmpt.begin(), tmpt.end(),
                 tmpt.begin(),
                 nume::Bignum::modulu_unary(num_->flag));

  std::vector<int> result(tmpr.size(), 0);

  std::transform(tmpt.begin(), tmpt.end(),
                 tmpr.begin(), result.begin(), std::plus<int>());

  std::swap(tmpt, result);

  normalize_();

  return *this;
}

nume::Bignum& nume::Bignum::operator-=(const nume::Bignum& rh)
{
  //////////////////////////////////////////////////////////////////////
  // 減算については、rhの符号を反転させて+=を行うことで実現可能である //
  //////////////////////////////////////////////////////////////////////

  nume::Bignum tmprh = rh;
  tmprh.num_->flag = reverse_flag_(rh.num_->flag);
  *this += tmprh;

  return *this;
}

nume::Bignum& nume::Bignum::operator*=(const nume::Bignum& rh)
{
  //////////////////////////////////////////////////////////////////////////
  // Bignum同士の乗算を行う。Bignum同士の乗算は、次のようにして行われる。 //
  //                                                                      //
  // 1. rhのdigitsの先頭と、自身のdigits全体をそれぞれ乗算する。          //
  // 2. rhのdigitsの先頭に対する乗算が完了したら、normalizeを行う。       //
  // 3. rhのdigitsの基準を一つ後ろに進め、結果をlhに加算する。            //
  // 4. rhの末尾に到達するまで上記を繰り返し、結果をlhとする。            //
  //////////////////////////////////////////////////////////////////////////

  Bignum b, c;
  std::vector<int> tmp(num_->digits.size(), 0);

  std::vector<int>::iterator it = rh.num_->digits.begin(), end = rh.num_->digits.end(),
    target = tmp.begin();

  int now_digits = 0;
  while (it != end) {
    std::advance(target, now_digits);
    std::transform(num_->digits.begin(), num_->digits.end(),
                   target, modulu_unary(*it++));

    c.num_->digits = tmp;
    b += c;

    tmp.push_back(0);
    target = tmp.begin(); now_digits++;
  }

  // 最終的にはこのようにして交換する。
  swap_(b);
  return *this;
}

nume::Bignum& nume::Bignum::operator/=(const nume::Bignum& rh)
{
  ////////////////////////////////////////////////
  // 与えられた右辺値で割ったデータを出力する。 //
  // 内部関数であるdivmodを利用する。           //
  ////////////////////////////////////////////////

  std::pair<nume::Bignum, nume::Bignum> p = divmod(rh);

  swap_(p.first);
  return *this;
}

nume::Bignum& nume::Bignum::operator%=(const nume::Bignum& rh)
{
  //////////////////////////////////////////////
  // 与えられた右辺値で割った余りを設定する。 //
  //////////////////////////////////////////////

  std::pair<nume::Bignum, nume::Bignum> p = divmod(rh);

  swap_(p.second);
  return *this;
}

nume::Bignum& nume::Bignum::operator++()
{
  *this += Bignum("1");
  return *this;
}

nume::Bignum& nume::Bignum::operator--()
{
  *this += Bignum("1");
  return *this;
}

nume::Bignum nume::Bignum::operator++(int n)
{
  Bignum b(*this);
  *this += Bignum("1");
  return b;
}

nume::Bignum nume::Bignum::operator--(int n)
{
  Bignum b(*this);
  *this -= Bignum("1");
  return b;
}

std::pair<nume::Bignum, nume::Bignum> nume::Bignum::divmod(const nume::Bignum& rh)
{
  ////////////////////////////////////////////
  // 与えられた右辺値による除算結果を返す。 //
  // firstが商、secondが余りとして返す。    //
  ////////////////////////////////////////////

  if (rh.num_->digits.empty() || rh == Bignum("0"))
  {
    throw BignumZeroDivideException("can't zero divide!");
  }

  Bignum lh = *this;
  std::pair<nume::Bignum, nume::Bignum> p;

  // 左辺値が右辺値未満になるまで減算を繰り返す。
  while (lh >= rh) {
    lh -= rh;
    ++p.first;
  }

  p.second = lh;

  return p;
}

////////////////////////////////////////////////////////
// 内部の機能を利用する、外部演算子オーバロードの定義 //
////////////////////////////////////////////////////////

bool nume::operator<=(const nume::Bignum& lh, const nume::Bignum& rh)
{
  return (lh < rh) || (lh == rh);
}

bool nume::operator!=(const nume::Bignum& lh, const nume::Bignum& rh)
{
  return !(lh == rh);
}

bool nume::operator>(const nume::Bignum& lh, const nume::Bignum& rh)
{
  return (rh < lh);
}

bool nume::operator>=(const nume::Bignum& lh, const nume::Bignum& rh)
{
  return (rh < lh) || (rh == lh);
}

nume::Bignum nume::operator+(const nume::Bignum& lh, const nume::Bignum& rh)
{
  nume::Bignum tmp(lh);
  tmp += rh;
  return tmp;
}

nume::Bignum nume::operator-(const nume::Bignum& lh, const nume::Bignum& rh)
{
  nume::Bignum tmp(lh);
  tmp -= rh;
  return tmp;
}

nume::Bignum nume::operator*(const nume::Bignum& lh, const nume::Bignum& rh)
{
  nume::Bignum tmp(lh);
  tmp *= rh;
  return tmp;
}

nume::Bignum nume::operator/(const nume::Bignum& lh, const nume::Bignum& rh)
{
  nume::Bignum tmp(lh);
  tmp /= rh;
  return tmp;
}

nume::Bignum nume::operator%(const nume::Bignum& lh, const nume::Bignum& rh)
{
  nume::Bignum tmp(lh);
  tmp %= rh;
  return tmp;
}

//////////////////////////////////////
// 内部で利用される構造体内部の定義 //
//////////////////////////////////////

nume::Bignum::convert_str::convert_str(int digits, std::stringstream& stream) :
  stream_(stream), digits_(digits)
{
}

void nume::Bignum::convert_str::operator()(int i)
{
  // 文字列をそのまま利用する。
  stream_ << std::setfill('0') << std::setw(digit_unit_size_) << i;
}
