#ifndef _BIGNUM_H_
#define _BIGNUM_H_

#include <utility>
#include <string>
#include "smart_ptr.h"
#include <vector>
#include <functional>

namespace utakata {

    namespace numeric {

        class BignumZeroDivideException : public std::exception
        {
        public:

            BignumZeroDivideException(std::string str) : str_(str) {}
            virtual ~BignumZeroDivideException() throw() {};
            const char* what() const throw() {
                return str_.c_str();
            }
        private:
            std::string str_;
        };

        class Bignum
        {
            //////////////////////////////////////////////////////////////////
            // 桁数がメモリのみに依存するような整数演算を行うためのクラス。 //
            // このクラス同士の演算である限り、正しい演算であることが       //
            // 保証されている。                                             //
            // また、このデータにおいては、10進数による計算のみが行われる。 //
            //////////////////////////////////////////////////////////////////

        private:

            template<class T>
            struct less_between : public std::binary_function<T, T, bool>
            {
                less_between(int flag) : flag_(flag) {
                }
                bool operator()(T x, T y) {
                    return flag_ > 0 ? x < y : x > y;
                }

                int flag_;
            };

            template<class T>
            struct greater_between : public std::binary_function<T, T, bool>
            {
                greater_between(int flag) : flag_(flag) {
                }
                bool operator()(T x, T y) {
                    return flag_ > 0 ? x > y : x < y;
                }

                int flag_;
            };

            struct convert_str : public std::unary_function<int, void>
            {
                convert_str(int digits, std::stringstream& stream);
                void operator()(int i);

                std::stringstream& stream_;
                const int digits_;
            };

            struct equal_zero : public std::unary_function<int, bool>
            {
                bool operator()(int t) {
                    return t != 0 ? true : false;
                }
            };

            struct not_equal_digits : public std::unary_function<char, bool>
            {
                bool operator()(char t) {
                    return !('0' <= t && t <= '9');
                }
            };

            struct modulu_unary : public std::unary_function<int, int>
            {
                modulu_unary(int f) : f_(f) {}
                int operator()(int l) {
                    return f_ * l;
                }

                int f_;
            };

        private:
            // 内部で利用する、各固定値
            enum {digit_unit_ = 100, digit_unit_size_ = 2};
            typedef std::pair<std::vector<int>::iterator, std::vector<int>::iterator> it_pair;

        public:
            Bignum();
            Bignum(const std::string& str);
            Bignum(const Bignum& rh);
            virtual ~Bignum() {}

            Bignum& operator=(const Bignum& rh);

            // 必須となるオペレータオーバロード
            bool operator<(const Bignum& rh) const;
            bool operator==(const Bignum& rh) const;

            // 各演算用のオペレータオーバロード
            Bignum& operator+=(const Bignum& rh);
            Bignum& operator-=(const Bignum& rh);
            Bignum& operator*=(const Bignum& rh);
            Bignum& operator/=(const Bignum& rh);
            Bignum& operator%=(const Bignum& rh);

            Bignum& operator++();
            Bignum operator++(int n);
            Bignum& operator--();
            Bignum operator--(int n);
            // 数値を文字列に変換する。
            std::string to_str() const;
            
        private:

            // 文字列を数値として認識する。与えられる文字列は完全な整
            // 数でなければならない。
            void parse_string_(const std::string& s);
            
            // 内部の数値を正規化する。
            void normalize_();

            // 別のbignumと内容の交換を行う。
            void swap_(Bignum& rh);

            // 与えられた数値の符号を返す。
            // reverse_flag_は、与えられた数値の符号の逆符号を返す。
            int to_flag_(int i) {return i > 0 ? 1 : -1;}
            int reverse_flag_(int i) {return to_flag_(i) > 0 ? -1 : 1;}

            // 現時点と現時点の次の値を元にして、桁毎の調整を
            // 行う。
            void normalize_one_(std::vector<int>::iterator it,
                                std::vector<int>::iterator fwd);

            // 与えられた右辺の値との商と余りを出力する。
            // firstが商、secondが余りを表す。
            std::pair<Bignum, Bignum> divmod(const Bignum& rh);

        private:
            // 内部で利用される構造体
            struct pImpl;
            smart_ptr<pImpl> num_;
        };

        ///////////////////////////////////////////////////////////////
        // 各比較演算子。<と==から以下はすべて導出することができる。 //
        ///////////////////////////////////////////////////////////////
        bool operator<=(const Bignum& lh, const Bignum& rh);
        bool operator!=(const Bignum& lh, const Bignum& rh);
        bool operator>(const Bignum& lh, const Bignum& rh);
        bool operator>=(const Bignum& lh, const Bignum& rh);

        ////////////////////////////////////
        // 各種計算を行うための二項演算子 //
        ////////////////////////////////////
        Bignum operator+(const Bignum& lh, const Bignum& rh);
        Bignum operator-(const Bignum& lh, const Bignum& rh);
        Bignum operator*(const Bignum& lh, const Bignum& rh);
        Bignum operator/(const Bignum& lh, const Bignum& rh);
        Bignum operator%(const Bignum& lh, const Bignum& rh);

    }
}

#endif /* _BIGNUM_H_ */
