#ifndef ___HGL___REF_DELETE___
#define ___HGL___REF_DELETE___

#include <utility>

/**
 ref_delete
 参照カウントを保持して、
 0になったときに、保持してあるオブジェクトを破棄するための
 クラス。
 smart_ptr中で、配列と非配列オブジェクトとを区別するために
 使用される。
*/
class ref_delete_base {
public:

    virtual ~ref_delete_base() {}

    ref_delete_base() : ref_(0), owner_(false),maxnum_(0),
                        objsize_(0) {}
    ref_delete_base(const ref_delete_base& base) :
        ref_(0), owner_(false),maxnum_(0), objsize_(0)
        {
            copy(base);
        }

    ref_delete_base& operator=(const ref_delete_base& base)
        {
            copy(base);
            return *this;
        }

    /**
     dec_ref
     参照カウントのデクリメント。
     これを実行したとき、参照カウントが0になったら、do_deleteを
     実行する。
     @access public
     @return bool true:削除済み false:未削除
    */
    bool dec_ref() {
        if (--ref_ <= 0) {
            if (owner_) {
                do_delete();
            }
            return true;
        }
        return false;
    }

    /**
     inc_ref
     参照カウントのインクリメント
     @access public
     @return void
    */
    void inc_ref() {
        ref_++;
    }

    /**
     getRef
     参照カウントの取得
     @access public
     @return int 参照カウント
    */
    int getRef() const {return ref_;}
    
    /**
     setRef
     参照カウントの設定
     @param int i 参照カウント
     @access public
     @return void
    */
    void setRef(int i) {ref_ = i;}

    /**
     setOwner
     所有権の設定
     @param bool b 所有権の有無
     @access public
     @return void
    */
    void setOwner(bool b) {owner_ = b;}
    
    /**
     getOwner
     所有権の取得
     @access public
     @return bool
    */
    bool getOwner() {return owner_;}

    /**
     getObjSize
     オブジェクトサイズの取得
     @access public
     @return int オブジェクトサイズ
    */
    int getObjSize() const {return objsize_;}
    
    /**
     setObjSize
     オブジェクトサイズの設定
     @param int i オブジェクトサイズ
     @access public
     @return void
    */
    void setObjSize(int i) {objsize_ = i;}

    /**
     getMaxNum
     最大インデックスの取得
     @access public
     @return int 最大インデックス
    */
    int getMaxNum() const {return maxnum_;}
    
    /**
     setMaxNum
     最大インデックスの設定
     @param int i 最大インデックス
     @access public
     @return void
    */
    void setMaxNum(int i) {maxnum_ = i;}

    /**
     do_delete
     参照カウントがゼロになった際に実行される。
     主に、deleteとdelete[]の違いを埋めるために使う
     @access protected
     @return void
    */
    virtual void do_delete() = 0;

    void copy(const ref_delete_base& base) {
        ref_ = base.ref_;
        owner_ = base.owner_;
        maxnum_ = base.maxnum_;
        objsize_ = base.objsize_;
    }

private:

    int ref_;         //参照カウント
    bool owner_;      //所有権
    int maxnum_;      //最大インデックス
    int objsize_;     //オブジェクトのサイズ

};

/**
 ref_noarray_object
 非配列オブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_noarray_object : public ref_delete_base {
public:

    ref_noarray_object(T* p) : obj_(p) {}
    ref_noarray_object(const ref_noarray_object<T>& o) : obj_(NULL)
        {
            copy(o);
            obj_ = o.obj_;
        }
    ref_noarray_object<T>& operator=(const ref_noarray_object<T>& o)
        {
            copy(o);
            obj_ = o.obj_;
            return *this;
        }

    virtual void do_delete() {
        if (obj_ != NULL) {
            delete obj_;
        }
    }

protected:

    T* obj_;

};

/**
 ref_array_object
 配列オブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_array_object : public ref_delete_base {
public:

    ref_array_object(T* p) : obj_(p) {}
    ref_array_object(const ref_array_object& o) : obj_(NULL)
        {
            copy(o);
            obj_ = o.obj_;
        }
    ref_array_object<T>& operator=(const ref_array_object& o)
        {
            copy(o);
            obj_ = o.obj_;
            return *this;
        }

    virtual void do_delete() {

        if (obj_ != NULL) {
            delete [] obj_;
        }
    }

protected:

    T* obj_;

};

/**
 ref_null_object
 所有権を持たないオブジェクトに対する、
 参照カウントを保持した削除クラス
*/
template<class T>
class ref_null_object : public ref_delete_base{
public:

    ref_null_object(T* p) : obj_(p) {}
    ref_null_object(const ref_null_object& o) : obj_(NULL)
        {
            copy(o);
            obj_ = o.obj_;
        }
    ref_null_object<T>& operator=(const ref_null_object& o)
        {
            copy(o);
            obj_ = o.obj_;
            return *this;
        }
    
    virtual void do_delete() {
    }

protected:

    T* obj_;

};

#endif
