#ifndef ___HGL___smart_ptr___
#define ___HGL___smart_ptr___

#include <iostream>
#include "ref_delete.h"

/**
   smart_ptr_base
   スマートポインタの基底クラス。
   特に何をするわけでもない。
*/
class smart_ptr_base {
public:
    virtual ~smart_ptr_base(){}
};

/**
   smart_ptr
   スマートポインタ。
   smart_ptr<CHoge> h(new CHoge);
   とやれば、後はコピーしても解体責任を委譲したりして、
   ちゃんとdeleteしてくれるようなクラス。
   また、配列もとれるようにして、配列であっても
   通常のポインタであっても、ちゃんとdeleteとdelete[]を
   分けてくれる。

   配列を生成する際には、

   smart_ptr<CHoge> h;
   h.addArray(true,10);
   のように生成するか、
   もしくは、CHogeの派生クラスSHogeを生成したりする場合は、
   h.addArray(new SHoge[10],true,10);

   のようにする。
*/
template<class T>
class smart_ptr // : public smart_ptr_base
{
public:

    //暗黙の型変換を行わない。
    //このコンストラクタでは、配列の生成は出来ないことになっている。
    template<class S> explicit 
    smart_ptr(S* p, bool owner = true, ref_delete_base* ref = NULL) : 
        obj_(NULL),ref_obj_(NULL),index_(0) {
        if (p != NULL) {
            init(p,owner,1, ref);
        }
    }

    //デフォルトコンストラクタ
    smart_ptr() : obj_(NULL),ref_obj_(NULL),index_(0) {
    }
    
    //コピーコンストラクタ
    smart_ptr(const smart_ptr<T>& p) : 
        obj_(NULL),ref_obj_(NULL),index_(0) {
        //参照カウントを増やしつつ、コピー。
        inc(p);
    }

    smart_ptr<T>& operator=(const smart_ptr<T>& obj) {
        // 自分にコピーされる前に、前のオブジェクトの参照カウントを
        // 減らしておく。
        dec();
        
        smart_ptr<T> tmp(obj);
        swap(tmp);
        return *this;
    }

    //参照数を減らすだけ。
    virtual ~smart_ptr() {dec();}

    //明示的に削除するための構文
    virtual void release() {dec();}

    //ここからは、ポインタの振りをするための仕掛け。
    T& operator*() const {return *get();}
    T* operator->() const {return get();}

    //擬似的な配列操作のため。
    smart_ptr<T>& operator++() {incIndex();return *this;}
    smart_ptr<T> operator++(int n) {
        smart_ptr<T> tmp(*this);
        incIndex();
        return tmp;
    }
    smart_ptr<T>& operator--() {decIndex();return *this;}
    smart_ptr<T> operator--(int n) {
        smart_ptr<T> tmp(*this);
        decIndex();
        return tmp;
    }

    // 同一かどうかをチェックする。
    bool operator==(const smart_ptr<T>& rh)
        {
            return obj_ == rh.obj_;
        }
    bool operator!=(const smart_ptr<T>& rh)
        {
            return !(*this == rh);
        }

    //配列からの取得のため。
    T& operator[](int n) const {return *get(n + 1);}

    /**
       get
       保持するメモリから、指定された場所のメモリを取得する。
       @param int n デフォルトは1
       @access public
       @return 指定された位置のテンプレートで指定されたオブジェクト
    */
    T* get(int n = 1) const {
        //配列、単一変数どちらにでも対応するため、
        //引数で場所を取る。
        

        if (ref_obj_ == NULL) {
            return NULL;
        }

        if (index_ + n > ref_obj_->getMaxNum() || index_ + n < 0) {
            return NULL;
        }

        //取得すべきメモリは、
        //オブジェクトの先頭アドレス+オブジェクトのサイズ*(インデックス+引数)
        //なお、この際に、index_ = 0かつ、n = 1である場合、
        //先頭アドレス+オブジェクトサイズということになり、どう考えてもアクセス違反になる。
        return (T*)((unsigned char*)obj_ + (ref_obj_->getObjSize() * (index_ + n - 1)));
    }

    //==============================================
    //疑似イテレーターとして使用するための関数群
    //==============================================

    /**
       isEnd
       インデックスが、限界に達していないかどうか
       ここで言う限界とは、MaxNumで、つまり添え字+1をさす。
       @access public
       @return bool 達している場合はtrue
    */
    bool isEnd() {
        if (ref_obj_ != NULL) {
            return ref_obj_->getMaxNum()-1 < index_?true:false;
        } else {
            return true;
        }
    }

    /**
       isBegin
       インデックスが、先頭であるかどうか
       @access public
       @return bool 先頭である場合はtrue
    */
    bool isBegin() {
        return index_ == 0?true:false;
    }

    /**
       オブジェクトがNULLかどうか
       @access public
       @return bool
    */
    bool isNull() const {
        return ref_obj_==NULL&&obj_==NULL?true:false;
    }

    // オブジェクトが単一の場所からのみ参照されているか。
    bool isOnlyRef() const {
        return ref_obj_->getRef() == 1 ? true : false;
    }
    
    /**
       end
       インデックスを、最後尾に位置させる
       @access public
       @return void
    */
    void end() {
        if (ref_obj_ != NULL) {
            index_ = ref_obj_->getMaxNum()-1;
        } else {
            index_ = 0;
        }
    }

    /**
       begin
       インデックスを、先頭に位置させる
       @access public
       @return void
    */
    void begin() {
        index_ = 0;
    }

    //=============================================
    //追加生成関数群
    //=============================================

    /**
     * @brief アップキャストを行う。SはTの派生クラスであること。
     * また、その特性上同一性チェックが困難であるので、こればかりは仕方がない。
     * @param const smart_ptr<S>& スマートポインタ
     * @return smart_ptr<T>& アップキャスト後の自身。
     */
    template<class S>
    smart_ptr<T>& upcast(const smart_ptr<S>& p) {
        
        if (reinterpret_cast<void*>(this) != 
            reinterpret_cast<void*>(const_cast<smart_ptr<S>*>(&p)) ) {
            //pがさすオブジェクトをこのオブジェクトもさすことになるので、
            //まずはdec
            dec();
            
            //あとはincとほぼ同じ処理になる。
            obj_ = p.getObj();
            ref_obj_ = p.getRefObj();
            index_ = p.getIndex();

            if (ref_obj_ != NULL) {
                ref_obj_->inc_ref();
            }
        }

        return *this;
    }

    /**
     * @brief 指定した型の追加生成構文
     * コンストラクタで何も指定しない場合、これを実行することで、再生成を行う。
     */
    void add() {
        dec();
        init(new T, true, 1);
    }

    /**
       add
       新しく生成させる。
       関数テンプレートを使ってはいるが、必ずTとおなじ、
       もしくはTの派生クラスであること。
       @param S* p 対象とするオブジェクト
       @param bool owner 所有権の有無
       @access public
       @return void
    */
    template<class S>
    void add(S* _p, bool owner = true) {
        dec();
        if (_p != NULL) {
            init(_p,owner,1);
        }
    }

    /**
     * @brief デリータを指定した追加生成構文。
     * これは配列では使用できない。
     */
    template<class S>
    void add(S* _p, bool owner, ref_delete_base* ref) {
        dec();
        if (_p != NULL && ref != NULL) {
            init(_p, owner, 1, ref);
        }
    }

    /**
       addArray
       配列オブジェクト生成関数
       現在のオブジェクトの型を、指定個数作成する場合。
       @param bool owner 所有権の有無
       @param int n 個数
       @access public
       @return void
    */
    void addArray(bool owner, int n) { 
        dec();
        if (n > 1) {
            init(new T[n], owner, n);
        } else if (n == 1) {
            init(new T, owner, 1);
        }
    }

    /**
       addArray
       配列オブジェクト生成関数
       こちらは、配列オブジェクトを生成する際に使用される関数。
       @param S p 対象とするオブジェクト
       @param bool owner 所有権の有無
       @param int n 個数
       @access public
       @return void
    */
    template<class S>
    void addArray(S _p, bool owner, int n) { 
        dec();
        if (n > 1) {
            init(new S[n], owner, n);
        }
    }


private:

    /**
       init
       スマートポインタの初期化補助関数
       渡されたオブジェクトを元にして、所有権、
       参照カウント、オブジェクトの個数を設定する。
       @param template _p 保持するべきオブジェクト
       @param bool owner 所有権の湯有無
       @param int n オブジェクトの個数
       @param ref_delete_base* ref デフォルトはNULL。外部からのコピーならば非NULL
       @access protected
       @return void
    */
    template<class S>
    void init(S* _p, bool owner, int n, ref_delete_base* ref = NULL) {
        
        index_ = 0;
        obj_ = _p;

        if (ref_obj_ != NULL) {
            //このとき、すでにdec_refとかはやってあることが前提。
            ref_obj_ = NULL;
        }

        if (ref != NULL) {
            ref_obj_ = ref;
        } else if (owner) {
            if (n == 1) {
                //非配列オブジェクト
                ref_obj_ = new ref_noarray_object<S>(_p);
            } else if (n > 1){
                ref_obj_ = new ref_array_object<S>(_p);
            }
        } else {
            //非所有権オブジェクト
            ref_obj_ = new ref_null_object<S>(_p);
        }

        //オブジェクトの情報を追加
        ref_obj_->setRef(1);
        ref_obj_->setMaxNum(n);
        ref_obj_->setObjSize(sizeof(S));
        ref_obj_->setOwner(owner);

    }

    /**
       inc
       渡されたsmart_ptrの参照数を一つ増やす。
       @param const smart_ptr<T>& p 参照カウントを増やすオブジェクト
       @access protected
       @return void
    */
    void inc(const smart_ptr<T>& p) {
        index_    = p.getIndex();
        ref_obj_   = p.getRefObj();
        obj_      = p.getObj();
        
        if (ref_obj_ != NULL) {
            ref_obj_->inc_ref();
        }
    }

    void swap(smart_ptr<T>& p)
    {
        // 対象となっているオブジェクトとデータを入れ替える。
        std::swap(index_, p.index_);
        std::swap(ref_obj_, p.ref_obj_);
        std::swap(obj_, p.obj_);
    }

    /**
       dec
       自分の分の参照カウントを減らす。
       実行されたら、ref_obj_およびobj_は
       NULLになる。
       @access protected
       @return void
    */
    void dec() {
        //dec_refは、オブジェクトを削除したかどうかを返す。
        //削除されている場合は、参照オブジェクトを削除しなければならない。
        if (ref_obj_ != NULL) {
            if (ref_obj_->dec_ref()) {
                delete ref_obj_;
            }
        }
        ref_obj_ = NULL;
        obj_ = NULL;
    }

    /**
       incIndex
       インデックスカウントを増やす
       @access protected
       @return void
    */
    void incIndex() {index_++;}

    /**
       decIndex
       インデックスカウントを減らす
       @access protected
       @return void
    */
    void decIndex() {index_--;}

    //========================================
    //コピーの際に使用するconst関数群
    //========================================

    /**
       getIndex
       現在のインデックスを返す
       @access public
       @return int 現在のインデックス
    */
    int getIndex() const {return index_;}

    /**
       getRefObj
       削除オブジェクトを返す
       @access public
       @return ref_delete_base* 削除オブジェクトのポインタ
    */
    ref_delete_base* getRefObj() const {return ref_obj_;}
    
    /**
       getObj
       オブジェクトを返す
       @access public
       @return T* オブジェクトのポインタ
    */
    T* getObj() const {return obj_;}

private:

    T* obj_;      //保持すべきオブジェクト。参照カウントオブジェクトと同一

    ref_delete_base* ref_obj_;     //参照カウントオブジェクト

    int index_;       //現在のインデックス

};

#endif
