#ifndef _LEXEME_DATA_H_
#define _LEXEME_DATA_H_

#include <string>
#include "src/common/smart_ptr.h"

namespace utakata {

namespace unicode {
class UniString;
class UniChar;
};

namespace lexeme {

struct StringData {
  // 文字列を保持するためのデータ型. たんにそのまま.
  smart_ptr<unicode::UniString> str;
};

struct NumberData {
  // 数値のデータの受け渡しを行うためのデータ型.
  // 数値データは, それぞれ実数部分と虚数部分, そして
  // 正確性と radix を定義する必要がある.
  // radix = 10 ではない場合, 常に exact = true として振る舞う.
  smart_ptr<unicode::UniString> real;
  smart_ptr<unicode::UniString> imagin;
  bool exact;
  int radix;
};

struct BooleanData {
  // Boolean を表すデータ. これでは bool 型の変数によって#t と#f を表現
  // することにする.
  bool boolean;
  smart_ptr<unicode::UniString> str;
};

struct CharactorData {
  // Charactor を構成するためのデータ.
  // この時点で、渡された文字列から数値に変換が完了しているものとする。
  smart_ptr<unicode::UniChar> spec;
};

struct SymbolData {
  // シンボルを構成するためのデータ.
  smart_ptr<unicode::UniString> id;
};

struct LexemeData {
  // リテラルとして存在するデータの一部のみを
  // 内包する.
  // このデータは, 実際には Lexeme と Object 間でのデータの受け渡しを
  // 行うためのインタフェースとなっている.
  // この上のlexemeによって、返される値が確定されると仮定するため、
  // タグ情報などは含まない。

  smart_ptr<StringData>    string;
  smart_ptr<NumberData>    number;
  smart_ptr<BooleanData>   boolean;
  smart_ptr<CharactorData> charactor;
  smart_ptr<SymbolData>    symbol;
};

// 各データ入りのLiteralDataを返すヘルパー関数。
smart_ptr<LexemeData> makeStringData(const unicode::UniString& str);
smart_ptr<LexemeData> makeNumberData(const unicode::UniString& real,
                                     const unicode::UniString& imagin,
                                     bool exact, bool radix);
smart_ptr<LexemeData> makeBooleanData(const unicode::UniString& str,
                                      bool boolean);
smart_ptr<LexemeData> makeCharactorData(const unicode::UniChar& ch);
smart_ptr<LexemeData> makeSymbolData(const unicode::UniString& str);
};
};

#endif /* _LEXEME_DATA_H_ */
