#include <vector>
#include <string>

#include "lexeme_impl.h"
#include "function_callback.h"
#include "unicode.h"

#include "lexeme_data.h"

using namespace utakata::lexeme;
using namespace utakata::unicode;

smart_ptr<ILexeme> utakata::lexeme::makeOpenParen()
{
    smart_ptr<UniString> tmp(new UniString(UniChar('('), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp, LexemeID::openParenthesis));
}

smart_ptr<ILexeme> utakata::lexeme::makeEOS()
{
    return smart_ptr<ILexeme>(new PureLexeme(smart_ptr<UniString>(), LexemeID::eos));
}

smart_ptr<ILexeme> utakata::lexeme::makeCloseParen()
{
    smart_ptr<UniString> tmp(new UniString(UniChar('('), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::closeParenthesis));
}

smart_ptr<ILexeme> utakata::lexeme::makeBackQuote()
{
    smart_ptr<UniString> tmp(new UniString(UniChar('`'), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::backquote));
}

smart_ptr<ILexeme> utakata::lexeme::makeQuote()
{
    smart_ptr<UniString> tmp(new UniString(UniChar('\''), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::quote));
}

smart_ptr<ILexeme> utakata::lexeme::makeDot()
{
    smart_ptr<UniString> tmp(new UniString(UniChar('.'), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::dot));
}

smart_ptr<ILexeme> utakata::lexeme::makeIdentifier(const utakata::unicode::UniString& str)
{
    return smart_ptr<ILexeme>(new Identifier(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeString(const utakata::unicode::UniString& str)
{
    return smart_ptr<ILexeme>(new String(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnquoteSplicing()
{
    smart_ptr<UniString> tmp(new UniString());
    *tmp += UniChar(',');
    *tmp += UniChar('@');
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unquoteSplicing));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnquote()
{
    smart_ptr<UniString> tmp(new UniString(UniChar(','), 1));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unquote));
}

smart_ptr<ILexeme> utakata::lexeme::makeByteVector()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#vu(")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::byteVector));
}

smart_ptr<ILexeme> utakata::lexeme::makeVector()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#(")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::vector));
}

smart_ptr<ILexeme> utakata::lexeme::makeSyntax()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#'")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::syntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeQuasiSyntax()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#`")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::quasiSyntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnsyntaxSplicing()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#,@")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unsyntaxSplicing));
}

smart_ptr<ILexeme> utakata::lexeme::makeUnsyntax()
{
    smart_ptr<UniString> tmp(new UniString(unicode::convert("#,")));
    return smart_ptr<ILexeme>(new PureLexeme(tmp,
                                             LexemeID::unsyntax));
}

smart_ptr<ILexeme> utakata::lexeme::makeCharactor(const utakata::unicode::UniChar& str)
{
    return smart_ptr<ILexeme>(new Charactor(str));
}

smart_ptr<ILexeme> utakata::lexeme::makeBoolean(smart_ptr<utakata::unicode::UniChar> ch)
{
    return smart_ptr<ILexeme>(new Boolean(ch));
}

smart_ptr<ILexeme> utakata::lexeme::makeNanImaginary(const utakata::unicode::UniString& str,
                                                     bool exact)
{
    return smart_ptr<ILexeme>(new Number(utakata::unicode::UniString(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeInfImaginary(const utakata::unicode::UniString& str,
                                                     bool exact)
{
    return smart_ptr<ILexeme>(new Number(utakata::unicode::UniString(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeImaginaryOnly(const utakata::unicode::UniString& str,
                                                      bool exact)
{

    return smart_ptr<ILexeme>(new Number(utakata::unicode::UniString(),
                                         str, exact, 10));
}

smart_ptr<ILexeme> utakata::lexeme::makeNumber(const utakata::unicode::UniString& real,
                                               const utakata::unicode::UniString& imagin,
                                               bool exact, int radix)
{
    return smart_ptr<ILexeme>(new Number(real, imagin, exact, radix));
}

////////////////
// PureLexeme //
////////////////

PureLexeme::PureLexeme(smart_ptr<UniString> str,
                       lexeme::LexemeID id) : str_(str), id_(id)
{}

const LexemeID PureLexeme::getID() const
{
    return id_;
}

smart_ptr<UniString> PureLexeme::toString() const
{
    return str_;
}

smart_ptr<utakata::lexeme::LexemeData> PureLexeme::getData() const
{
    return smart_ptr<utakata::lexeme::LexemeData>();
}

////////////////
// Identifier //
////////////////

Identifier::Identifier(const utakata::unicode::UniString& str) :
    str_(new utakata::unicode::UniString(str))
{
}

const utakata::lexeme::LexemeID Identifier::getID() const
{
    return utakata::lexeme::LexemeID::identifier;
}

smart_ptr<utakata::unicode::UniString> Identifier::toString() const
{
    return str_;
}

smart_ptr<utakata::lexeme::LexemeData> Identifier::getData() const
{
    return makeSymbolData(*str_);
}

/////////////
// String  //
/////////////

String::String(const utakata::unicode::UniString& str) :
    str_(new utakata::unicode::UniString(str))
{
}

const utakata::lexeme::LexemeID String::getID() const
{
    return utakata::lexeme::LexemeID::string;
}

smart_ptr<utakata::unicode::UniString> String::toString() const
{
    return str_;
}

smart_ptr<utakata::lexeme::LexemeData> String::getData() const
{
    return makeStringData(*str_);
}

////////////
// Number //
////////////

Number::Number(const UniString& real,
               const UniString& imagin,
               bool exact, int radix) :
    data_(new LexemeData)
{
    data_ = makeNumberData(real, imagin, exact, radix);
}

const utakata::lexeme::LexemeID Number::getID() const
{
    return utakata::lexeme::LexemeID::number;
}

smart_ptr<utakata::unicode::UniString> Number::toString() const
{
    return smart_ptr<UniString>(new UniString(*data_->number->real + *data_->number->imagin));
}

smart_ptr<utakata::lexeme::LexemeData> Number::getData() const
{
    return data_;
}

///////////////
// Charactor //
///////////////

Charactor::Charactor(const UniChar& ch) : ch_(new UniChar(ch))
{
}

const utakata::lexeme::LexemeID Charactor::getID() const
{
    return utakata::lexeme::LexemeID::charactor;
}

smart_ptr<utakata::unicode::UniString> Charactor::toString() const
{
    smart_ptr<utakata::unicode::UniString> c(new utakata::unicode::UniString(*ch_, 1));
    return c;
}

smart_ptr<utakata::lexeme::LexemeData> Charactor::getData() const
{
    return makeCharactorData(*ch_);
}

/////////////
// Boolean //
/////////////

Boolean::Boolean(smart_ptr<UniChar> ch) : bool_(new UniString())
{
    *bool_ += '#';
    *bool_ += *ch;
}

const utakata::lexeme::LexemeID Boolean::getID() const
{
    return utakata::lexeme::LexemeID::boolean;
}

smart_ptr<utakata::unicode::UniString> Boolean::toString() const
{
    return bool_;
}

smart_ptr<utakata::lexeme::LexemeData> Boolean::getData() const
{
    return makeBooleanData(*bool_, true);
}
