#ifndef _LEXEME_IMPL_H_
#define _LEXEME_IMPL_H_

#include <string>

#include "src/lexeme.h"
#include "src/common/smart_ptr.h"
#include "src/common/function_callback.h"
#include "src/lexeme_id.h"

// lexeme の様々な実装を定義する.
namespace utakata {

namespace unicode {
class UniString;
class UniChar;
};

namespace lexeme {
class LexemeID;
class LexemeData;

// 各 lexeme を smart_ptr にして取得する.
smart_ptr<ILexeme> makeOpenParen();
smart_ptr<ILexeme> makeEOS();
smart_ptr<ILexeme> makeCloseParen();
smart_ptr<ILexeme> makeBackQuote();
smart_ptr<ILexeme> makeQuote();
smart_ptr<ILexeme> makeDot();
smart_ptr<ILexeme> makeUnquoteSplicing();
smart_ptr<ILexeme> makeUnquote();
smart_ptr<ILexeme> makeSyntax();
smart_ptr<ILexeme> makeQuasiSyntax();
smart_ptr<ILexeme> makeUnsyntaxSplicing();
smart_ptr<ILexeme> makeUnsyntax();
smart_ptr<ILexeme> makeByteVector();
smart_ptr<ILexeme> makeVector();
smart_ptr<ILexeme> makeIdentifier(const utakata::unicode::UniString& str);
smart_ptr<ILexeme> makeString(const utakata::unicode::UniString& str);
smart_ptr<ILexeme> makeNumber(const utakata::unicode::UniString& real,
                              const utakata::unicode::UniString& imaginary,
                              bool exact, int radix);
smart_ptr<ILexeme> makeCharactor(const utakata::unicode::UniChar& str);
smart_ptr<ILexeme> makeNanImaginary(const utakata::unicode::UniString& str,
                                    bool exact);
smart_ptr<ILexeme> makeInfImaginary(const utakata::unicode::UniString& str,
                                    bool exact);
smart_ptr<ILexeme> makeImaginaryOnly(const utakata::unicode::UniString& str,
                                     bool exact);
smart_ptr<ILexeme> makeBoolean(smart_ptr<utakata::unicode::UniChar> ch);


////////////////////////////////////////////
// 実際に利用する lexeme は以下に実装する. //
////////////////////////////////////////////

class PureLexeme : public ILexeme {
  // LiteralDataを必要としないデータに対して処理を行うためのクラス。
  // 文字列と該当するLexemeIDを渡すことで実現する。
 public:
  PureLexeme(smart_ptr<utakata::unicode::UniString> str,
             lexeme::LexemeID id);

  virtual ~PureLexeme() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;

 private:

  smart_ptr<utakata::unicode::UniString> str_;
  lexeme::LexemeID id_;
};

////////////
// String //
////////////

class String : public ILexeme {
 public:

  explicit String(const utakata::unicode::UniString& ident);
  virtual ~String() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;

 private:

  smart_ptr<utakata::unicode::UniString> str_;
};

////////////////
// Identifier //
////////////////

class Identifier : public ILexeme {
 public:

  explicit Identifier(const utakata::unicode::UniString& ident);
  virtual ~Identifier() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;

 private:

  smart_ptr<utakata::unicode::UniString> str_;
};

////////////
// Number //
////////////

class Number : public ILexeme {
 public:

  Number(const unicode::UniString& real, const unicode::UniString& imagin,
         bool exact, int radix);
  virtual ~Number() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;
 private:

  smart_ptr<utakata::lexeme::LexemeData> data_;
};

///////////////
// Charactor //
///////////////

class Charactor : public ILexeme {
 public:

  explicit Charactor(const unicode::UniChar& ch);
  virtual ~Charactor() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;

 private:

  smart_ptr<utakata::unicode::UniChar> ch_;
};

/////////////
// Boolean //
/////////////

class Boolean : public ILexeme {
 public:

  explicit Boolean(smart_ptr<unicode::UniChar> ch);
  virtual ~Boolean() {}

  const lexeme::LexemeID getID() const;
  smart_ptr<utakata::unicode::UniString> toString() const;
  smart_ptr<utakata::lexeme::LexemeData> getData() const;

 private:
  smart_ptr<unicode::UniString> bool_;
};
};                                      // end of namespace `lexeme`
};                                      // end of namespace `utakata`

#endif /* _LEXEME_IMPL_H_ */
