#ifndef _NUMBER_H_
#define _NUMBER_H_

// utakataにおけるschemeの数値集合を実装する。
// R6RS準拠では、数値の継承関係を正確に実装しなければならない。
// utakataでは、正確数の最大値についてはメモリの容量にのみ依存するが、
// 非正確数の最大/最小についてはIEEEの仕様に準拠することとする。

///////////////////////////////////////////////////////////////////////////////
// utakataにおける数値集合を作成する際の処理については以下のような形になる。 //
//                                                                           //
// real    -- 小数点が含まれている場合、小数点が0であるならば正確数          //
//         -- 小数点が含まれていない場合、正確数                             //
// imagin  -- 小数点が含まれている場合、小数点が0であるならば正確数          //
//         -- 小数点が含まれていない場合、正確数                             //
// radix   -- 2,8,10,16のどれかが指定されるか、もしくは指定されないか。      //
//            指定されている場合、それぞれ変換する必要があるが、10以外では   //
//            小数点は扱わないため、常に正確数となる。                       //
// exact   -- #eが設定されている場合に処理される。#eの場合は、強制的に正確数 //
//            となるが、real,imaginが非正確数の場合、結局非正確数になる。    //
///////////////////////////////////////////////////////////////////////////////

// 上記のように、型と正確性は直交するため、内部で保持される値は、doubleと正確数の
// 双方が保持されるようになっていて、非正確数の場合には内部的な値がdoubleに変換される。
// この部分のインターフェースについては、別に記載する。

namespace utakata {

namespace numeric {

class Number {
  /**
     Schemeの数階層の最上位となるクラス。
     このクラスは実際にはインスタンス化されず、
     型として存在しているだけになる。
  */
 public:
  Number() {}
  virtual ~Number() {}
};

class Complex : public Number {
  /**
     複素数を扱うクラス。
     複素数は、有理数を含む全ての実数を表現することが可能であるため、
     Numberの直下として働く。
     非正確数については、IEEEの制限範囲内であり、正確数については、メモリにのみ
     依存する。
  */
 public:
  Complex() {}
  virtual ~Complex() {}
};

class Real : public Complex {
  /**
     schemeの実数を表現するクラス。
     有理数と整数、および無理数を表現できる。
  */
 public:
  Real() {}
  virtual ~Real() {}
};

class Rational : public Real {
  /**
     schemeの有理数を表現するクラス。
     有理数と整数を表現することができる。
     有理数とは、整数aとbの除算で表現できる小数となる。
     整数同士の除算で表現できない場合、その数は無理数となり、実数となる。
  */
 public:
  Rational() {}
  virtual ~Rational() {}
};

class Integer : public Rational {
  /**
     schemeの整数を表現するクラス。
     整数はその通りであり、概要でいえば、小数点が存在しない数になる。
     正確な整数は、原則としてメモリのみに桁数が依存される。
     非正確数の場合には、あくまでdoubleに依存する。
  */
 public:
  Integer() {}
  virtual ~Integer() {}
};
};
};

#endif /* _NUMBER_H_ */
