#ifndef _NUMBER_LEXER_H_
#define _NUMBER_LEXER_H_

#include <iostream>
#include <sstream>

#include "src/unicode.h"
#include "src/common/smart_ptr.h"
#include "src/reader.h"
#include "src/lexeme.h"
#include "src/sublexer.h"

namespace utakata {

namespace sublexer {

namespace number {

/////////////////////////////////////////////////////////////////
// <number>がそれぞれのradixに基づいて解析する必要があるため、 //
// それぞれに対して特化できるようなテンプレートを用意する。    //
/////////////////////////////////////////////////////////////////

template<int NUM> class UReal;
template<int NUM>
class Number {
 private:
  const int INIT;
  const int REAL;     // <real>非終端記号を示す。これは決して虚数にはならない。
  const int UREAL;    // <ureal>非終端記号を示す。
  const int NAN_IMAGINARY;
  const int INF_IMAGINARY;
  const int IMAGINARY_ONLY;
  const int FLAG;
  const int END;
  const int IMAGINARY;
 public:
  Number() :
      INIT(1), REAL(2), UREAL(3), NAN_IMAGINARY(4), INF_IMAGINARY(5),
      IMAGINARY_ONLY(8), FLAG(9), END(10), IMAGINARY(11),
      exact_(false), realnum_(new unicode::UniString()),
      imaginary_(new unicode::UniString()) {
  }

  virtual ~Number() {}

  smart_ptr<lexeme::ILexeme> Lex(smart_ptr<reader::StreamReader> stream,
                                 smart_ptr<ISubLexer>& next,
                                 const unicode::UniString& str,
                                 bool exactness) {
    // 全体的な流れは同一であり、基数の違いによる
    // 数値の違い程度しか問題は発生しない。
    exact_ = exactness;
    lexer_delimiter::Normal nor;
    unicode::UniString number = str;
    while (true) {
      // デリミタまで読んでしまう。
      if (stream->isEOF()) {
        throw utakata::sublexer::LexException(pos_, "illegal number format");
      }

      unicode::UniChar ch(stream->read());
      if (nor(ch)) {
        stream->unget(ch.GetRawCode());
        break;
      }

      number += ch;
    }

    // 読出したnumberを解析してNumberオブジェクトとして返す。
    pos_ = stream->pos();
    return lexNumber_(number);
  }

 private:

  // 現在値を前の値に退避し、現在値をnextで更新する。
  void SetStatus(int next, int *now, int *prev) { // NOLINT
    *prev = *now;
    *now = next;
  }

  // 現在値をnextで更新する。
  void SetStatus(int next, int* now) { // NOLINT
    *now = next;
  }

  smart_ptr<lexeme::ILexeme> lexNumber_(const unicode::UniString& number) {
    // デリミタまでを取得したデータを、実際の数値オブジェクトとして解釈させる。
    // 解釈できなかった場合、字句構文エラーとして扱う。
    unicode::UniString::const_iterator begin = number.begin(),
        end = number.end();

    int status = INIT, prevstatus = INIT;
    bool imaginary = false;

    while (begin != end) {
      if (status == INIT) {
        // 非終端記号で言えば<complex>に当たる。
        // 全ての整数は原則として +-0iの虚数と等しいため、そのようになる。
        SetStatus(LexComplex(begin, end), status, prevstatus);

      } else if (status == REAL) {
        // 前回の状態によって、チェックが変更になる。
        SetStatus(real_(begin, end), status, prevstatus);

      } else if (status == UREAL) {
        // 基本的にはそのまま設定するのみで構わない。ただしこの後に続く
        // 可能性があるためそのチェックも同時に行う。
        bool b = UReal<NUM>(pos_).Lex(begin, end);
        if (b) {
          if (!imaginary && begin != end) {
            // まだ続きがある。
            realnum_->Insert(realnum_->begin(), number.begin(), begin);
            SetStatus(IMAGINARY, status);

          } else {
            // 上記以外の場合、基本的に終了しているはずなので返る。
            SetStatus(END, status);
            realnum_->Insert(realnum_->begin(), number.begin(), begin);
            return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
          }
        }
      } else if (status == IMAGINARY) {
        // 虚数部分の判定となる。
        SetStatus(status, prevstatus, flag_(begin, end));

      } else if (status == END) {
        // 終了ということなので、完全に完了したことにする。
        return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);

      } else if (status == FLAG) {
        // 虚数部分の先頭符号を示す。
        unicode::UniString s;
        s.Insert(s.begin(), begin, end);

        if (s.size() == 1) {
          throw utakata::sublexer::LexException(
              pos_, "imaginary parts must constain `i' at end");
        } else if (s[s.size() - 1].GetRawCode() != 'i') {
          throw utakata::sublexer::LexException(
              pos_, "imaginary parts must constain `i' at end");
        }

        if (s[1].GetRawCode() == 'i' && s.size() == 2) {
          // 末尾がiの場合、これがそのまま文字列となる。
          *imaginary_ = s;
          return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);

        } else {
          // +i,-i以外の場合は、限定領域がrealである必要がある。
          unicode::UniString s2;
          s2.Insert(s2.begin(), s.begin() + 1, s.end() - 1);
          if (s2 == unicode::convert("nan.0") ||
              s2 == unicode::convert("inf.0")) {
            *imaginary_ = s;
            return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);

          } else {
            // urealである必要がある。
            utakata::unicode::UniString::const_iterator begin = s2.begin(),
                end = s2.end();
            UReal<NUM>(pos_).Lex(begin, end);
            // 例外が発生しなければureal_である。
            *imaginary_ = s;
            return lexeme::makeNumber(*realnum_, *imaginary_, exact_, NUM);
          }
        }
      } else if (status == NAN_IMAGINARY) {
        return lexeme::makeNanImaginary(number, exact_);

      } else if (status == INF_IMAGINARY) {
        return lexeme::makeInfImaginary(number, exact_);

      } else if (status == IMAGINARY_ONLY) {
        return lexeme::makeImaginaryOnly(number, exact_);
      }
    }
    return smart_ptr<lexeme::ILexeme>();
  }

  // 複素数の要素の中で、終端記号であるものであるかどうかを判定する。

  int LexComplex(const unicode::UniString::const_iterator& begin,
                 const unicode::UniString::const_iterator& end) {
    // 一文字単位で調べていく。checker_に該当するデータの場合には、
    unicode::UniString temp_digits;
    temp_digits.Insert(temp_digits.begin(), begin, end);

    int 
    if (temp_digits == unicode::Convert("-nan.0i") ||
        temp_digits == unicode::Convert("+nan.0i")) {
      // 虚数部が正/負の非数として登録する。
      *imaginary_ = temp_digits;
      return NAN_IMAGINARY;
    } else if (temp_digits == unicode::Convert("-inf.0i") ||
               temp_digits == unicode::Convert("+inf.0i")) {
      // 虚数部が正/負の無限大として処理する。
      *imaginary_ = temp_digits;
      return INF_IMAGINARY;
    } else if (temp_digits == unicode::Convert("-i") ||
               temp_digits == unicode::Convert("+i")) {
      // 0+1i,0-1iにそれぞれマッピングされる。
      *imaginary_ = temp_digits;
      return IMAGINARY_ONLY;
    }

    // 上記以外の場合、必ず先頭はREALである必要がある。

    return REAL;
  }

  int real_(unicode::UniString::const_iterator& begin,
            const unicode::UniString::const_iterator& end) {
    // 実数の表現でなければならない。
    // 先頭一文字が符号表現か、符号無し表現で
    // あること。
    if (begin->GetRawCode() == '+' || begin->GetRawCode() == '-') {
      // 符号表現である場合には、以降の符号
      // 表現までを調べる。
      return innerReal_(begin, end, true);
    } else {
      return innerReal_(begin, end, false);
    }
  }

  int innerReal_(unicode::UniString::const_iterator& begin,
                 const unicode::UniString::const_iterator& end,
                 bool flag) {
    // realの符号有無に関係の無い、共通部分の分離。
    if (flag) {
      ++begin;
    }
    unicode::UniString s;
    unicode::UniString::const_iterator save = begin;
    while (begin != end && (begin->GetRawCode() != '+' ||
                            begin->GetRawCode() != '-')) {
      ++begin;
    }

    s.Insert(s.begin(), save, begin);
    // とりあえず抜けた段階で、実際に抽出
    // してみたデータを調べてみる。
    if (s == unicode::Convert("nan.0") || s == unicode::Convert("inf.0")) {
      // 非数/無限大として出力する。
      *realnum_ = s;
      return REAL;
    } else {
      // <uinteger>,<uinteger>/<uinteger>,<decimal 10>のいずれか
      // である必要がある。
      begin = save;
      return UREAL;
    }
  }


  int flag_(unicode::UniString::const_iterator& begin,
            const unicode::UniString::const_iterator& end) {
    // 虚数の先頭に必ず必要な符号が存在するかどうかを返す。
    if (begin->GetRawCode() == '+' || begin->GetRawCode() == '-') {
      return FLAG;
    }
    // これが無い場合、特に問題は発生しない。
    if (begin == end) {
      return END;
    }

    throw utakata::sublexer::LexException(pos_, "illegal number format : invalid flag previous `imaginary'");
  }

  // チェックを行うための関数オブジェクト。
  bool exact_;
  smart_ptr<unicode::UniString> realnum_;
  smart_ptr<unicode::UniString> imaginary_;
  size_t pos_;
};

////////////////////////////////
// <uinteger>非終端記号の実装 //
////////////////////////////////

template<int NUM>
class UInteger {
  // 指定された基数の値であるかどうかを調べる。
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  UInteger() {}
  virtual ~UInteger(){}

  bool operator()(iter it) {
    return true;
  }
};

template<>
class UInteger<2> {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  bool operator()(iter it) {
    if (it->GetRawCode() == '0' || it->GetRawCode() == '1') {
      return true;
    }
    return false;
  }
};

template<>
class UInteger<8> {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  bool operator()(iter it) {
    if (it->GetRawCode() == '0' || it->GetRawCode() == '1' ||
        it->GetRawCode() == '2' || it->GetRawCode() == '3' ||
        it->GetRawCode() == '4' || it->GetRawCode() == '5' ||
        it->GetRawCode() == '6' || it->GetRawCode() == '7') {
      return true;
    }
    return false;
  }
};

template<>
class UInteger<10> {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  bool operator()(iter it) {
    if (it->GetRawCode() == '0' || it->GetRawCode() == '1' ||
        it->GetRawCode() == '2' || it->GetRawCode() == '3' ||
        it->GetRawCode() == '4' || it->GetRawCode() == '5' ||
        it->GetRawCode() == '6' || it->GetRawCode() == '7' ||
        it->GetRawCode() == '8' || it->GetRawCode() == '9') {
      return true;
    }
    return false;
  }
};

template<>
class UInteger<16> {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  bool operator()(iter it) {
    if (it->GetRawCode() == '0' || it->GetRawCode() == '1' ||
        it->GetRawCode() == '2' || it->GetRawCode() == '3' ||
        it->GetRawCode() == '4' || it->GetRawCode() == '5' ||
        it->GetRawCode() == '6' || it->GetRawCode() == '7' ||
        it->GetRawCode() == '8' || it->GetRawCode() == '9' ||
        it->GetRawCode() == 'A' || it->GetRawCode() == 'B' ||
        it->GetRawCode() == 'C' || it->GetRawCode() == 'D' ||
        it->GetRawCode() == 'E' || it->GetRawCode() == 'F' ||
        it->GetRawCode() == 'a' || it->GetRawCode() == 'b' ||
        it->GetRawCode() == 'c' || it->GetRawCode() == 'd' ||
        it->GetRawCode() == 'e' || it->GetRawCode() == 'f') {
      return true;
    }
    return false;
  }
};

/////////////////////////////////
// <ureal>非終端記号解析の実装 //
/////////////////////////////////

template<int NUM>
class UReal {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  UReal(size_t pos) : pos_(pos) {}
  virtual ~UReal() {}

  int Lex(iter& begin, const iter end) {
    // <ureal>に以降の文字列が該当するかどうかを調べる。

    iter save = begin;

    for (; begin != end && checker_(begin); ++begin);

    if (begin == end) {
      // すべて数値の間に入っている場合、beginからendまでの全体が
      // real/imaginaryのいずれかとして扱われる。
      return true;
    } else {
      // すべてが数値の間ではなかった場合、その文字を調べる。
      // ここに至って、10進数とそれ以外とでチェックに相違点が存在する。
      if (begin->GetRawCode() == '/') {
        // 次もuintegerである必要がある。
        ++begin;
        // マッチする間飛ばす。
        for (; begin != end && checker_(begin); ++begin);
        return true;
      } else {
        throw utakata::sublexer::LexException(pos_, "must have digit at least one after `/'");
      }
    }
  }

 private:

  UInteger<NUM> checker_;
  size_t pos_;
};

template<>
class UReal<10> {
  typedef utakata::unicode::UniString::const_iterator iter;
 public:
  UReal(size_t pos) : pos_(pos) {}
  virtual ~UReal() {}

  bool Lex(iter& begin, const iter& end) {
    // <ureal>に以降の文字列が該当するかどうかを調べる。

    iter save = begin;
    if (begin->GetRawCode() == '.') {
      // 先頭が.の場合、そのまま進む。
      for (; begin != end && checker_(begin); ++begin);
      GetSuffix_(begin, end);
      return true;
    } else {
      // 先頭が.じゃない場合、一度数値を読み出しておく。
      for (; begin != end && checker_(begin); ++begin);

      if (begin == end) {
        //   throw utakata::sublexer::LexException(pos_, "illegal number format : not Real");
        return true;
      }

      if (begin->GetRawCode() == '.') {
        // 以降の文字がsuffixとかであるかどうかで調べる。
        ++begin;
        for (; begin != end && checker_(begin); ++begin);
      }
      GetSuffix_(begin, end);
      return true;
    }
  }

 private:

  bool GetSuffix_(iter& begin, const iter& end)
  {
    // 次の文字がsuffixであるかどうかを調べる。
    if (begin->GetRawCode() == 'e' || begin->GetRawCode() == 'E' ||
        begin->GetRawCode() == 's' || begin->GetRawCode() == 'S' ||
        begin->GetRawCode() == 'f' || begin->GetRawCode() == 'F' ||
        begin->GetRawCode() == 'd' || begin->GetRawCode() == 'D' ||
        begin->GetRawCode() == 'l' || begin->GetRawCode() == 'L') {
      ++begin;
      if (begin->GetRawCode() == '+' || begin->GetRawCode() == '-') {
        ++begin;
        if (checker_(begin)) {
          for (; begin != end && checker_(begin); ++begin);
          return true;
        } else {
          throw utakata::sublexer::LexException(pos_, "suffix nessesary after at least one number");
        }
      }
    }

    return true;
  }

  UInteger<10> checker_;
  size_t pos_;
};
};
};
};

#endif /* _NUMBER_LEXER_H_ */
