#include "numeric.h"

using namespace utakata;

numeric::NumericValue::NumericValue(const unicode ::UniString& val,
                                    bool exact) :
    inexact_(), exact(), exactness_(exact)
{
    // 渡されたexactnessを元にして、正確数と非正確数のいずれかを作成する。
}

numeric::NumericValue::NumericValue(const numeric::ExactValue& val) :
    inexact_(), exact(new numeric::ExactValue(val)), exactness_(true)
{
    // 正確数が確定する。
}

numeric::NumericValue::NumericValue(const numeric::InexactValue& val) :
    inexact_(new numeric::InexactValue(val)), exact(), exactness_(false)
{
    // 非正確数を確定する。
}

numeric::NumericValue::NumericValue(const numeric::NumericValue& val) :
    inexact_(), exact(), exactness_(false)
{
    // 渡された値とのswapを行うのみ。
    swap(val);
}

void numeric::NumericValue::swap(const numeric::NumericValue& val)
{
    // 渡されたNumericValueのswapをおこなう
    exactness_ = val.exactness_;

    if (exactness_)
    {
        exact_.add(new numeric::InexactValue(*val.exact_));
        inexact_.release();
    }
    else
    {
        inexact_.add(new numeric::InexactValue(*val.inexact_));
        exact_.release();
    }
}

numeric::NumericValue& numeric::NumericValue::operator=(const numeric::NumericValue& val)
{
    // 新規に一つ作成して、それをswapさせる。
    numeric::NumericValue tmp(val);
    swap(tmp);
    return *this;
}

const numeric::InexactValue& numeric::NumericValue::inexact() const
{
    // 内部で保持している非正確数を返す。
    // 内部が正確数の場合、これを実行すると例外を発生させる。
    return *inexact_;
}

const numeric::ExactValue& numeric::NumericValue::exact() const
{
    // 内部で保持している正確数を返す。
    // 内部が非正確数の場合、これを実行すると例外を発生させる。
    return *exact_;
}

numeric::NumericValue& numeric::NumericValue::operator=(const numeric::NumericValue& val)
{
    // 代入する。
    // ここでの代入は、あくまで保持されている値をコピーすることにある。
    numeric::NumericValue tmp(val);
    swap(tmp);
    return *this;
}

numeric::NumericValue& numeric::NumericValue::operator-=(const numeric::NumericValue& rh)
{
    // 内部の値との加算を行う。
    // 加算の際には、非正確数が伝播するように減算を行う。
    numeric::NumericValue tmp(*this - rh);
    swap(tmp);
    return *this;
}

numeric::NumericValue& numeric::NumericValue::operator+=(const numeric::NumericValue& rh)
{
    // 内部の値との加算を行う。
    // 加算の際には、非正確数が伝播するように加算を行う。
    numeric::NumericValue tmp(*this + rh);
    swap(tmp);
    return *this;
}

numeric::NumericValue& numeric::NumericValue::operator*=(const numeric::NumericValue& rh)
{
    // 内部の値との加算を行う。
    // 加算の際には、非正確数が伝播するように乗算を行う。
    numeric::NumericValue tmp(*this * rh);
    swap(tmp);
    return *this;
}

numeric::NumericValue& numeric::NumericValue::operator/=(const numeric::NumericValue& rh)
{
    // 内部の値との加算を行う。
    // 加算の際には、非正確数が伝播するように除算を行う。
    numeric::NumericValue tmp(*this / rh);
    swap(tmp);
    return *this;
}



//================================================================================

numeric::NumericValue numeric::operator+(const numeric::NumericValue& lh,
                                         const numeric::NumericValue& rh)
{
    if (!rh.isExact() && lh.isExact())
    {
        // 左辺値を非正確数として計算を行う。
        InexactValue val = numeric::exact2inexact(lh.exact());
        numeric::NumericValue ret(val + rh.inexact());
        return ret;
    }
    else if (rh.isExact() && !lh.isExact())
    {
        // 右辺値を非正確数として計算を行う。
        InexactValue i = numeric::inexact2exact(rh.exact());
        numeric::NumericValue ret(lh.inexact() + i);
        return ret;
    }
    else if (isExact())
    {
        // ここまで来た時点で、どちらも正確数である。
        // 両方が正確数なので、算出結果を渡せばよい。
        return numeric::NumericValue(lh.exact() + rh.exact());
    }
    else
    {
        // 非正確数同士の加算である。
        return numeric::NumericValue(lh.inexact() + rh.inexact());
    }
}

numeric::NumericValue numeric::operator-(const numeric::NumericValue& lh,
                                         const numeric::NumericValue& rh)
{
    if (!rh.isExact() && lh.isExact())
    {
        // 左辺値を非正確数として計算を行う。
        InexactValue val = numeric::exact2inexact(lh.exact());
        numeric::NumericValue ret(val - rh.inexact());
        return ret;
    }
    else if (rh.isExact() && !lh.isExact())
    {
        // 右辺値を非正確数として計算を行う。
        InexactValue i = numeric::inexact2exact(rh.exact());
        numeric::NumericValue ret(lh.inexact() - i);
        return ret;
    }
    else if (isExact())
    {
        // ここまで来た時点で、どちらも正確数である。
        // 両方が正確数なので、算出結果を渡せばよい。
        return numeric::NumericValue(lh.exact() - rh.exact());
    }
    else
    {
        // 非正確数同士の加算である。
        return numeric::NumericValue(lh.inexact() - rh.inexact());
    }
}

numeric::NumericValue numeric::operator*(const numeric::NumericValue& lh,
                                         const numeric::NumericValue& rh)
{
    if (!rh.isExact() && lh.isExact())
    {
        // 左辺値を非正確数として計算を行う。
        InexactValue val = numeric::exact2inexact(lh.exact());
        numeric::NumericValue ret(val * rh.inexact());
        return ret;
    }
    else if (rh.isExact() && !lh.isExact())
    {
        // 右辺値を非正確数として計算を行う。
        InexactValue i = numeric::inexact2exact(rh.exact());
        numeric::NumericValue ret(lh.inexact() * i);
        return ret;
    }
    else if (isExact())
    {
        // ここまで来た時点で、どちらも正確数である。
        // 両方が正確数なので、算出結果を渡せばよい。
        return numeric::NumericValue(lh.exact() * rh.exact());
    }
    else
    {
        // 非正確数同士の加算である。
        return numeric::NumericValue(lh.inexact() * rh.inexact());
    }
}

numeric::NumericValue numeric::operator/(const numeric::NumericValue& lh,
                                         const numeric::NumericValue& rh)
{
    if (!rh.isExact() && lh.isExact())
    {
        // 左辺値を非正確数として計算を行う。
        InexactValue val = numeric::exact2inexact(lh.exact());
        numeric::NumericValue ret(val / rh.inexact());
        return ret;
    }
    else if (rh.isExact() && !lh.isExact())
    {
        // 右辺値を非正確数として計算を行う。
        InexactValue i = numeric::inexact2exact(rh.exact());
        numeric::NumericValue ret(lh.inexact() / i);
        return ret;
    }
    else if (isExact())
    {
        // ここまで来た時点で、どちらも正確数である。
        // 両方が正確数なので、算出結果を渡せばよい。
        return numeric::NumericValue(lh.exact() / rh.exact());
    }
    else
    {
        // 非正確数同士の加算である。
        return numeric::NumericValue(lh.inexact() / rh.inexact());
    }
}

