#ifndef _NUMERIC_H_
#define _NUMERIC_H_

namespace utakata {

    namespace unicode {
        class UniString;
    };

    namespace numeric {
    
        class NumericValue
        {
            // 正確数と非正確数の両方をまとめて保持し、相互運用を行うことができるようにする。
        public:
            // 大元は文字列から作成させることができる。
            NumericValue(const unicode::UniString& val, bool exact);
            // それぞれのexactnessに確定させる。
            NumericValue(const numeric::ExactValue& e);
            NumericValue(const numeric::InexactValue& i);

            // 単純なコピーで十分。
            NumericValue(const NumericValue& v);
            virtual ~NumericValue() {}

            bool isExact() const {return exactness_;}

            const numeric::InexactValue& inexact() const;
            const numeric::ExactValue& exact() const;

            // 代入のオペレータオーバロード。
            NumericValue& operator=(const NumericValue& v);

            // 各加減乗除の定義。
            NumericValue& operator+=(const NumericValue& rh);
            NumericValue& operator-=(const NumericValue& rh);
            NumericValue& operator*=(const NumericValue& rh);
            NumericValue& operator/=(const NumericValue& rh);

        private:

            void swap(const NumericValue& val);

            // 正確数と非正確数の両方を保持する可能性がある。
            smart_ptr<numeric::InexactValue> inexact_;
            smart_ptr<numeric::ExactValue> exact_;

            // 正確数を保持しているのか非正確数を保持しているのか。
            bool exactness_;

        };

        // 各種比較関数。
        // 比較の時点で、片方が非正確数なら、一方も非正確数で計算
        // するようにする。

        // 各計算はほぼ同じ判定をおこなうため、判定をまとめて実行する。
        template<class F>
        numeric::NumericValue numeric::NumericOperation(const numeric::NumericValue& lh,
                                                        const numeric::NumericValue& rh,
                                                        F f)
        {
            if (!rh.isExact() && lh.isExact())
            {
                // 左辺値を非正確数として計算を行う。
                InexactValue val = numeric::exact2inexact(lh.exact());
                numeric::NumericValue ret(f(val, rh.inexact()));
                return ret;
            }
            else if (rh.isExact() && !lh.isExact())
            {
                // 右辺値を非正確数として計算を行う。
                InexactValue i = numeric::inexact2exact(rh.exact());
                numeric::NumericValue ret(f(lh.inexact(), i));
                return ret;
            }
            else if (isExact())
            {
                // ここまで来た時点で、どちらも正確数である。
                // 両方が正確数なので、算出結果を渡せばよい。
                return numeric::NumericValue(f(lh.exact(), rh.exact()));
            }
            else
            {
                // 非正確数同士の加算である。
                return numeric::NumericValue(f(lh.inexact(), rh.inexact()));
            }
        }

        bool operator==(const NumericValue& lh, const NumericValue& rh);
        bool operator>(const NumericValue& lh, const NumericValue& rh);
        bool operator<(const NumericValue& lh, const NumericValue& rh);

        // 既存だけでなんとか算出できるような演算子のオーバーロード
        bool operator!=(const NumericValue& lh, const NumericValue& rh) {
            return !(lh == rh);
        }

        bool operator<=(const NumericValue& lh, const NumericValue& lh) {
            return !(lh > rh);
        }
        bool operator>=(const NumericValue& lh, const NumericValue& rh) {
            return !(lh < rh);
        }

        // 通常の加減乗除は外部で定義する。
        NumericValue operator+(const NumericValue& lh, const NumericValue& rh);
        NumericValue operator-(const NumericValue& lh, const NumericValue& rh);
        NumericValue operator*(const NumericValue& lh, const NumericValue& rh);
        NumericValue operator/(const NumericValue& lh, const NumericValue& rh);
        NumericValue operator%(const NumericValue& lh, const NumericValue& rh);

        // 正確数と非正確数の変換はここで行えるようにする。
        NumericValue toExact(const NumericValue& v);
        NumericValue toInexact(const NumericValue& v);
    
    };

};

#endif /* _NUMERIC_H_ */
