#ifndef _OBJECT_H_
#define _OBJECT_H_

#include "smart_ptr.h"
#include "data_structure.h"

namespace utakata {

    namespace type {
        class Type;
    };

    namespace data {

        class Object
        {

            // インタプリタ中でオブジェクトとして扱われる全てのオブジェクト
            // となる。
            // 構文データも含め、全てのデータを取り扱う。
            // 全てのオブジェクトは、このオブジェクトに単純に「束縛」されるのみと
            // なり、実装はその先のオブジェクトに完全に移譲されている。
            // このオブジェクトは、そのままschemeでの変数となり、このオブジェクト
            // が保持するのは、「変更可能性」と「オブジェクト格納領域」のみ
            // となる。

        public:

            Object(bool changeable, const smart_ptr<DataSpace>& space);
            Object(const data::Object& o);

            Object& operator=(const Object& o);
            virtual ~Object(){}
            // 変更可能なオブジェクトであるかどうかを返す。
            virtual bool enableChange() const;

            // 実体のデータが格納されている領域を返す。
            virtual const smart_ptr<DataSpace> getSpace() const;
            virtual void setSpace(const smart_ptr<DataSpace>& space);

        private:

            void swap(const Object& o);

            smart_ptr<DataSpace> space_;
            bool changeable_;
        };

        class NullObject : public Object
        {
            // ObjectのNull実装。このオブジェクトは、ObjectがNULL = 値
            // を保持していないことを保証する。
            // 通常のObjectでは、getSpace().isNullは成立しないが、このオブジェクトは
            // 必ず成立する。
            // このオブジェクトは、Nilと=であるconsに設定される。
            // nilオブジェクトと、NullObjectは全く違う。
        public:

            NullObject();
            virtual ~NullObject() {}

            // 変更可能なオブジェクトであるかどうかを返す。
            virtual bool enableChange() const {return false;}

            // 実体のデータが格納されている領域を返す。
            virtual smart_ptr<DataSpace> getSpace() {return smart_ptr<DataSpace>();}
            virtual void setSpace(const smart_ptr<DataSpace>& space) {}
        };

        // 渡されたオブジェクトが、同一のTypeであるかどうかを返す。
        // type::toTypeと合わせて利用するイメージで利用される。
        bool instanceof(const data::Object& o, const type::Type& type);

    };

};

#endif /* _OBJECT_H_ */
