#include <sstream>
#include <stack>

#include "parser.h"
#include "lexeme.h"
#include "lexeme_data.h"
#include "lexer.h"
#include "lexeme_id.h"
#include "datum_id.h"
#include "reader.h"
#include "object.h"
#include "gc.h"
#include "primitive.h"
#include "primitive_class.h"
#include "primitive_util.h"
#include "data_castor.h"

using namespace utakata;
using namespace utakata::parser;
using namespace utakata::lexeme;
using namespace utakata::interpreter;

parser::DatumException::DatumException(std::string str) : str_()
{
    // エラーメッセージを定義する。
    std::stringstream ss;
    ss << "datum error ! -- message : [" << str << "]" << std::endl;
    str_ = ss.str();
}

const char* parser::DatumException::what() const throw()
{
    return str_.c_str();
}

////////////////////////////
// Parser Implementations //
///////////////////////////

parser::Parser::Parser(smart_ptr<lexer::Lexer> l) : lexer_(l),
                                                    handler_()
{
    // 初期状態用ハンドラの設定
    std::pair<PARSERSTATUS, smart_ptr<IParserHandler> > p;
    p.first = PS_INIT;
    p.second.add(new InitHandler);
    handler_.insert(p);

    // リスト内部用ハンドラの設定
    p.first = PS_LIST;
    p.second.add(new ListHandler);
    handler_.insert(p);

    // .内用ハンドラの設定
    p.first = PS_END_DOT;
    p.second.add(new DotEndHandler);
    handler_.insert(p);

    // abbreviation 用ハンドラの設定
    p.first = PS_ABBREVIATION;
    p.second.add(new AbbrevHandler);
    handler_.insert(p);

    // error 用のハンドラの設定
    p.first = PS_ERROR;
    p.second.add(new ErrorHandler);
    handler_.insert(p);
}

smart_ptr<data::Object> parser::Parser::parse(smart_ptr<reader::StreamReader>& strm,
                                              gc::GarbageCollector& gc)
{
    HandlerData data(gc);
    data.status.push(PS_INIT);

    while(!data.status.empty()) {

        if(data.status.top() != PS_ERROR)
        {
            if (data.chain)
            {
                data.chain = false;
            }
            else
            {
                data.lexm = lexer_->lex(strm);
            }
            
            if(data.lexm.isNull() ||
               data.lexm->getID() == lexeme::LexemeID::eos)
            {
                break;
            }
            // 基本的にすべてハンドラに任せる。
            handler_[data.status.top()]->exec(data);
        }
        else
        {
            // ERROR だったら、強制的にハンドラを回してループから抜ける。
            handler_[data.status.top()]->exec(data);
        }
    }

    // stackのtopにはmainが殘ったままなので、data.stackのtopを追加してやる。
    smart_ptr<data::Object> ret(new data::Object(data.stack.top()));
    return ret;
}

bool parser::isAbbrev(smart_ptr<lexeme::ILexeme> l)
{
    // abbreviation かどうかを判別する。
    switch(l->getID().toEnum())
    {
    case lexeme::LexemeID::BACKQUOTE:        return true;
    case lexeme::LexemeID::QUOTE:            return true;
    case lexeme::LexemeID::UNQUOTE:          return true;
    case lexeme::LexemeID::UNQUOTESPLICING:  return true;
    case lexeme::LexemeID::SYNTAX:           return true;
    case lexeme::LexemeID::QUASISYNTAX:      return true;
    case lexeme::LexemeID::UNSYNTAX:         return true;
    case lexeme::LexemeID::UNSYNTAXSPLICING: return true;
    default:
        return false;
    }
}

interpreter::IObject* parser::makePrimitiveFromLexeme(smart_ptr<lexeme::ILexeme> l)
{
    // 各lexemeに対応した固定lexemeを返す。
    interpreter::IObject* o = NULL;
    switch (l->getID().toEnum())
    {
    case lexeme::LexemeID::STRING:
        return util::makeString(l->getData()->string->str);
    case lexeme::LexemeID::CHARACTOR:
        return util::makeCharactor(l->getData()->charactor->spec);
    case lexeme::LexemeID::IDENTIFIER:
        return util::makeSymbol(l->getData()->symbol->id);
    case lexeme::LexemeID::BOOLEAN:
        return util::makeBoolean(l->getData()->boolean->boolean);
    case lexeme::LexemeID::NUMBER:
    {
        lexeme::NumberData& t = *(l->getData()->number);
        return util::makeNumber(t.real, t.imagin, t.exact,
                                t.radix);
    }
    case lexeme::LexemeID::BACKQUOTE:
        return util::Abbrev<primitive::Quasiquote>::make();
    case lexeme::LexemeID::QUOTE:
        return util::Abbrev<primitive::Quote>::make();
    case lexeme::LexemeID::UNQUOTE:
        return util::Abbrev<primitive::Unquote>::make();
    case lexeme::LexemeID::UNQUOTESPLICING:
        return util::Abbrev<primitive::UnquoteSplicing>::make();
    case lexeme::LexemeID::SYNTAX:
        return util::Abbrev<primitive::Syntax>::make();
    case lexeme::LexemeID::QUASISYNTAX:
        return util::Abbrev<primitive::Quasisyntax>::make();
    case lexeme::LexemeID::UNSYNTAX:
        return util::Abbrev<primitive::Unsyntax>::make();
    case lexeme::LexemeID::UNSYNTAXSPLICING:
        return util::Abbrev<primitive::UnsyntaxSplicing>::make();
    default:
        break;
    }

    return o;
}

/////////////////////
// parser handlers //
/////////////////////

bool InitHandler::exec_(HandlerData& data)
{
    data.status.pop();
    // 初期状態の解釈処理を行う。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 開き括弧である場合、compoundsに新しくpushして返る。
        data.compounds.push(std::vector<data::Object>());
        data.status.push(PS_LIST);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::string ||
            data.lexm->getID() == lexeme::LexemeID::number ||
            data.lexm->getID() == lexeme::LexemeID::identifier ||
            data.lexm->getID() == lexeme::LexemeID::charactor ||
            data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        // それぞれの場合、単純にstackに各オブジェクトをpushする。

        data::Object o = data::Object(false,
                                      data.gc.add(makePrimitiveFromLexeme(data.lexm)));
        if (data.compounds.empty())
        {
            data.stack.push(o);
        }
        else
        {
            data.compounds.top().push_back(o);
        }
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation である場合には、 全体をリストに展開するため、carまでは
        // 設定された状態にする。
        std::vector<data::Object> v;
        v.push_back(data::Object(false,
                                 data.gc.add(makePrimitiveFromLexeme(data.lexm))));

        data.compounds.push(v);

        // status stack に、 PS_ABBREVIATIONとPS_INITを積む。
        data.status.push(PS_ABBREVIATION);
        data.status.push(PS_INIT);
    }
    else
    {
        throw DatumException("datum の開始記号ではありません");
    }

    return true;
}

// List 内部処理のハンドリング

bool ListHandler::exec_(HandlerData& data)
{
    // list の開始部分である場合。
    if(data.lexm->getID() == lexeme::LexemeID::openParenthesis)
    {
        // 更にcompoundsに追加して積む。
        data.compounds.push(std::vector<data::Object>());
        data.status.push(PS_LIST);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::closeParenthesis)
    {
        // 原則として、閉括弧が来た時点で終了とするだけでよい。
        // 現在のtopにあるcompoundsが、データであるとされているため、
        // とりあえずtopにあるvectorをConsの繋がりに変換する。

        std::vector<data::Object> v = data.compounds.top(); data.compounds.pop();
        // vectorをConsへと変換するには、次のようにして後ろからやっていく
        // 方式が一番面倒がない。
        // 1. cdrをnil、carをvectorの末尾であるデータとしたconsを作成する。
        // 2. vectorの次のデータをcar、前のconsをcdrとしたconsを作成する。
        // 3. 先頭まで続ける。

        if (v.size() == 0)
        {
            // サイズが0の場合には、nilを設定する。
            data::Object o = data::Object(false,
                                          data.gc.add(util::makeNil()));
            pushList(data, o);
        }
        else
        {
            // 後ろから前に向かっていく。
            std::vector<data::Object>::reverse_iterator begin = v.rbegin(),
                end = v.rend();
            // ちょっと面倒すぎるなこれ。
            data::Object o = data::Object(
                false, data.gc.add(
                    util::makeCons(*begin,
                                   data::Object(false, data.gc.add(util::makeNil()))))
                );
            ++begin;
            for (; begin != end; ++begin)
            {
                // 順次oに再設定していく。
                o = data::Object(
                    false, data.gc.add(
                        util::makeCons(*begin, o)));
            }
            pushList(data, o);
        }

        data.status.pop();
    }
    else if (data.lexm->getID() == lexeme::LexemeID::string ||
             data.lexm->getID() == lexeme::LexemeID::number ||
             data.lexm->getID() == lexeme::LexemeID::identifier ||
             data.lexm->getID() == lexeme::LexemeID::charactor ||
             data.lexm->getID() == lexeme::LexemeID::boolean)
    {
        // データスタックの頂上にあるvectorに追加する。
        data.compounds.top().push_back(
            data::Object(
                false, data.gc.add(makePrimitiveFromLexeme(data.lexm))));
    }
    else if(isAbbrev(data.lexm))
    {
        // abbreviation は、結論としてはリストに展開される。carは常にabbreviationである
        // とすることで問題は無いので、作成したconsを、topに置くことで実現する。

        std::vector<data::Object> v;
        v.push_back(data::Object(false,
                                 data.gc.add(makePrimitiveFromLexeme(data.lexm))));

        data.compounds.push(v);

        // status stack に、 PS_ABBREVIATION と PS_INIT を積む。
        data.status.push(PS_ABBREVIATION);
        data.status.push(PS_INIT);
    }
    else if(data.lexm->getID() == lexeme::LexemeID::dot)
    {
        // .が来た場合で、topのvectorがemptyだったらどうしようもなくなるので終了。
        
        if (data.compounds.top().empty())
        {
            data.status.push(PS_ERROR);
            data.error = "necessary one datum previous dot";
        }
        else
        {
            // この時点で、終わり方が確定したため、別のハンドラに任せる。
            // ただし、compoundの方は特に変化は必要が無いので、そちらはそのまま。
            data.status.pop();
            data.status.push(PS_END_DOT);
            data.status.push(PS_INIT);
        }
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    return true;
}

void ListHandler::pushList(HandlerData& data, data::Object& o)
{
    if (!data.compounds.empty())
    {
        data.compounds.top().push_back(o);
    }
    else
    {
        data.stack.push(o);
    }
}

bool DotEndHandler::exec_(HandlerData& data)
{
    // .後の) を待つ部分のハンドラ。
    // このハンドラでは、) 以外が来た場合にはすべて構文エラーとなる。
    data.status.pop();
    if(data.lexm->getID() == lexeme::LexemeID::closeParenthesis)
    {
        // stackの頂点とcompoundsのトップを取得する。
        std::vector<data::Object> v = data.compounds.top(); data.compounds.pop();
        // vectorをConsへと変換するには、次のようにして後ろからやっていく
        // 方式が一番面倒がない。
        // 1. cdrをstackのtop、carをvectorの末尾であるデータとしたconsを作成する。
        // 2. vectorの次のデータをcar、前のconsをcdrとしたconsを作成する。
        // 3. 先頭まで続ける。

        {
            // 後ろから前に向かっていく。
            std::vector<data::Object>::reverse_iterator begin = v.rbegin(),
                end = v.rend();
            // ちょっと面倒すぎるなこれ。
            // cdrにcompoundsの末尾を、carにその前を設定する。
            data::Object o = data::Object(
                false, data.gc.add(
                        util::makeCons(begin[1], begin[0]))
                );
            begin += 2;
            for (; begin != end; ++begin)
            {
                // 順次oに再設定していく。
                o = data::Object(
                    false, data.gc.add(
                        util::makeCons(*begin, o)));
            }

            if (!data.compounds.empty())
            {
                data.compounds.top().push_back(o);
            }
            else
            {
                data.stack.push(o);
            }
        }

        // ここまでは、すべてDataSpace内部のDataEntityを対象に処理しているため、
        // Objectで正しく受け渡しができていれば問題は発生しない。
    }
    else
    {
        // 上記以外は許容しない。
        data.status.push(PS_ERROR);
    }

    return true;
}

// Abbreviation 処理のハンドリング。

bool AbbrevHandler::exec_(HandlerData& data)
{
    // abbreviationの場合、すでにPS_INITを事前に積んであるため、
    // 一つ分のdatumが来ていることが確定している。
    // すでにabbreviation分のObjectはcompoundのtopに積んであるため、
    // consを作成して、compoundかstackのtopに積むだけ。

    std::vector<data::Object> v = data.compounds.top(); data.compounds.pop();

    if (v.size() != 2)
    {
        data.status.push(PS_ERROR);
    }
    else
    {
        // まだcompoundsが存在している場合には、これを返すようにする。
        data::Object c = data::Object(
            false, data.gc.add(util::makeCons(
                                   v[1],
                                   data::Object(false, data.gc.add(util::makeNil()))
                                   )));
        data::Object o =
            data::Object(false,
                         data.gc.add(util::makeCons(v[0], c)));
        if (!data.compounds.empty())
        {
            data.compounds.top().push_back(o);
        }
        else
        {
            data.stack.push(o);
        }
    }

    data.status.pop();
    data.chain = true;

    return true;
}

// syntax error のハンドリング

bool ErrorHandler::exec_(HandlerData& data)
{
    // 基本的にはエラーのハンドリングを行うだけ。
    // 一応発生した前後の位置と発生した lexeme くらいは表示したいかな。
    std::stringstream ss;
    ss << "syntax error：" << data.error;
    throw DatumException(ss.str());

    return true;
}
