#ifndef _PARSER_H_
#define _PARSER_H_

#include <string>
#include <vector>
#include <map>

#include "smart_ptr.h"
#include "parser_handler.h"
#include "parser_status.h"

namespace utakata {

    namespace syntax {
        class DatumID;
    };

    namespace interpreter {
        class Object;
        class IObject;
    };

    namespace lexer {
        class Lexer;
    };

    namespace reader {
        class StreamReader;
    };

    namespace gc {
        class GarbageCollector;
    };

    namespace parser {
    
        class DatumException : public std::exception
        {
        public:
            DatumException(std::string str);
            virtual ~DatumException() throw() {}
            const char* what() const throw();

        private:

            std::string str_;
        };

        class Parser
        {
            // lexerを受け取り、lexerが保持する全体の解析を行いつつ、
            // 実行を行う。
            // 実行は、基本的に全体を読み込み、全体を保持した状態で行う。
            
        public:
            
            Parser(smart_ptr<lexer::Lexer> l);
            virtual ~Parser() {}

            // 受け取ったlexerからの字句の取得を行い、解析を行なう。
            // 成功した場合には、datum一つを表す状態を返す。
            // 致命的なエラーが発生した場合には例外を返す。
            smart_ptr<data::Object> parse(smart_ptr<reader::StreamReader>& strm,
                                          gc::GarbageCollector& gc);

        private:
            smart_ptr<lexer::Lexer> lexer_;
            // 各PARSERSTATUSに関連づけられたParserHandlerのmap
            std::map<PARSERSTATUS, smart_ptr<IParserHandler> > handler_;

        };

        // 渡したlexemeがAbbrevationであるかどうかを調べる。
        bool isAbbrev(smart_ptr<lexeme::ILexeme> l);

        // 渡されたlexemeに対応する各lexeme datumを生成して返す。
        interpreter::IObject* makePrimitiveFromLexeme(smart_ptr<lexeme::ILexeme> l);

        // 各ハンドラの定義。

        class InitHandler : public IParserHandler
        {
            // PS_INITに対応するハンドラ
        public:
            InitHandler() {}
            virtual ~InitHandler() {}

        private:

            bool exec_(HandlerData &data);

        };

        // <list>に対応するものだけは、carとcdrでハンドラが分離している。
        class ListHandler : public IParserHandler
        {
            // PS_LIST_CARである間に対応するハンドラ
        public:
            ListHandler() {}
            virtual ~ListHandler() {}

        private:

            bool exec_(HandlerData& data);

            // 現在のハンドラデータの内容から、pushすべきスタックを
            // 決定してpushする。
            void pushList(HandlerData& data, data::Object& o);

        };

        // .後のdatumに対応するためのハンドラ
        class DotEndHandler : public IParserHandler
        {
            // PS_END_DOTに対応するハンドラ。
        public:
            DotEndHandler() {}
            virtual ~DotEndHandler() {}

        private:

            bool exec_(HandlerData& data);

        };


        class AbbrevHandler : public IParserHandler
        {
            // PS_ABBREVIATIONに対応するハンドラ
        public:
            AbbrevHandler() {}
            virtual ~AbbrevHandler() {}

        private:

            bool exec_(HandlerData& data);

        };

        class ErrorHandler : public IParserHandler
        {
            // PS_ERRORをハンドリングする。実際にはエラーを通告する
            // のみ。
        public:
            ErrorHandler() {}
            virtual ~ErrorHandler() {}
            
        private:

            bool exec_(HandlerData& data);

        };

    };

};

#endif /* _PARSER_h_ */
