#ifndef _PRIMITIVE_H_
#define _PRIMITIVE_H_

#include <vector>
#include "smart_ptr.h"
#include "base_object.h"

namespace utakata {

    namespace unicode {
        class UniChar;
        class UniString;
    };

    namespace data {
        class Object;
    };

    namespace numeric {
        class Number;
    };

    namespace interpreter {

        class IClass;
        namespace primitive {

            // data_structureに基づいた、各リテラルの実体を実装する。
            // ここでいうプリミティブとは、以下を指す。
            // ・文字列 string
            // ・数値   number
            // ・文字   charactor
            // ・真偽値 boolean
            // ・シンボル symbol
            // ・バイトベクタ bytevector
            // ・ベクタ vector
            // ・コンス cons
            // これらはすべて、単独のオブジェクトとして存在することが可能であり、
            // またリテラルの記法として存在しているものとなる。

            ///////////////////////////
            // implementation String //
            ///////////////////////////

            class String : public IObject
            {
                // scheme中での文字列を指す。
            public:
                String() : data_() {}
                String(const smart_ptr<unicode::UniString>& data);
                virtual ~String() {}

                unicode::UniString& getData();

                smart_ptr<interpreter::IClass> getClass();

            private:

                smart_ptr<unicode::UniString> data_;
            };

            //////////////////////////////
            // implementation Charactor //
            //////////////////////////////

            class Charactor : public IObject
            {
                // scheme中でのunicode文字を表す。
                // 一文字のみを表現するため、UniCharが利用される。
            public:
                Charactor() : data_() {}
                Charactor(const smart_ptr<unicode::UniString>& data);
                virtual ~Charactor() {}

                unicode::UniChar& getData();
                smart_ptr<interpreter::IClass> getClass();

            private:

                smart_ptr<unicode::UniChar> data_;
            };


            //////////////////////////////
            // implementation Number    //
            //////////////////////////////

            class Number : public IObject
            {
                // scheme中における数値オブジェクトを作成し、保持する。
                // 数値オブジェクトは、渡されたデータのコピーか、もしくは
                // 渡された数値に該当するNumberオブジェクトを生成して保持する。
            public:
                Number(const smart_ptr<unicode::UniString>& real,
                       const smart_ptr<unicode::UniString>& imagin,
                       bool exact, int radix);
                virtual ~Number() {}

                smart_ptr<interpreter::IClass> getClass();

            private:

                // 内部で保持する数値実体。
                smart_ptr<numeric::Number> num_ent_;

            };

            ////////////////////////////
            // implementation Boolean //
            ////////////////////////////

            class Boolean : public IObject
            {
                // scheme中でのBooleanを表す。
                // Booleanは、true/falseのいずれかを一意に示すものであり、
                // 単純にtrueかfalseかのみを返すようになっている。
            public:
                Boolean() : boolean_(false) {}
                Boolean(bool b);
                virtual ~Boolean() {}

                bool getData() const;
                smart_ptr<interpreter::IClass> getClass();
            private:
                bool boolean_;
            };


            ///////////////////////////
            // implementation Symbol //
            ///////////////////////////

            class Symbol : public IObject
            {
                // シンボルを表現する。シンボルは実際には単純な文字列として
                // 保持されるが、処理の際に特別に処理される点が異なっている。
            public:
                Symbol() : symbol_() {}
                Symbol(const smart_ptr<unicode::UniString>& sym);
                virtual ~Symbol() {}

                unicode::UniString& getData();
                smart_ptr<interpreter::IClass> getClass();

            private:

                smart_ptr<unicode::UniString> symbol_;
            };

            /////////////////////////
            // implementation Cons //
            /////////////////////////

            class Cons : public IObject
            {
                // リストの基本表現であるcons。listはconsの連結のみで表現
                // される。
                // car,cdrの表現もこれを用いておこなわれる。
            public:
                Cons(const data::Object& car,
                     const data::Object& cdr);
                virtual ~Cons() {}

                data::Object& car();
                data::Object& cdr();

                void setCar(const data::Object& car);
                void setCdr(const data::Object& cdr);

                bool isNil() const;

                smart_ptr<interpreter::IClass> getClass();

            private:

                smart_ptr<data::Object> car_;
                smart_ptr<data::Object> cdr_;
            };

            ///////////////////////////////
            // implementation Bytevector //
            ///////////////////////////////

            class ByteVector : public IObject
            {
                // バイトベクタを表現する。内部的には完全にvectorと
                // 一致するが、型の相違を表現するためにわざと型を分岐させている。
            public:
                ByteVector() : vector_() {}
                ByteVector(const std::vector<data::Object>& vec);
                virtual ~ByteVector() {}

                std::vector<data::Object>& getData();
                smart_ptr<interpreter::IClass> getClass();

            private:

                std::vector<data::Object> vector_;
            };


            ///////////////////////////
            // implementation Vector //
            ///////////////////////////

            class Vector : public IObject
            {
                // 単純な配列を表現する。Bytevectorと内部的には一致するが、
                // 同様に型を区別するために分離させている。
            public:
                Vector() : vector_() {}
                Vector(const std::vector<data::Object>& vec);
                virtual ~Vector() {}

                std::vector<data::Object>& getData();
                smart_ptr<interpreter::IClass> getClass();

            private:

                std::vector<data::Object> vector_;
            };

            ///////////////////////////
            // implementation Nil    //
            ///////////////////////////

            class Nil : public IObject
            {
                // car, vectorのいずれにもオブジェクトが設定されていない
                // オブジェクトを指す。
            public:
                Nil() {}
                virtual ~Nil() {}

                smart_ptr<interpreter::IClass> getClass();

            };

            ///////////////////////////
            // implementation Quote  //
            ///////////////////////////

            class Quote : public IObject
            {
                // 以降に渡されたdatumをそのまま返すためのクォート記号。
            public:
                Quote() {}
                virtual ~Quote() {}

                smart_ptr<interpreter::IClass> getClass();
            };

            class Quasiquote : public IObject
            {
            public:
                Quasiquote() {}
                virtual ~Quasiquote() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class Unquote : public IObject
            {
            public:
                Unquote() {}
                virtual ~Unquote() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class UnquoteSplicing : public IObject
            {
            public:
                UnquoteSplicing() {}
                virtual ~UnquoteSplicing() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class Syntax : public IObject
            {
            public:
                Syntax() {}
                virtual ~Syntax() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class Quasisyntax : public IObject
            {
            public:
                Quasisyntax() {}
                virtual ~Quasisyntax() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class Unsyntax : public IObject
            {
            public:
                Unsyntax() {}
                virtual ~Unsyntax() {}
                smart_ptr<interpreter::IClass> getClass();
            };

            class UnsyntaxSplicing : public IObject
            {
            public:
                UnsyntaxSplicing() {}
                virtual ~UnsyntaxSplicing() {}
                smart_ptr<interpreter::IClass> getClass();
            };

        }; // end of namespace primitive
    };

};

#endif /* _PRIMITIVE_H_ */
