#include <assert.h>

#include "object.h"
#include "primitive.h"
#include "primitive_util.h"
#include "primitive_class.h"
#include "type.h"
#include "data_castor.h"

using namespace utakata;
using namespace utakata::interpreter;

IObject* interpreter::util::makeCons(const data::Object& car,
                                     const data::Object& cdr)
{
    return new interpreter::primitive::Cons(car, cdr);
}

IObject* interpreter::util::makeNil()
{
    return new interpreter::primitive::Nil();
}

IObject* interpreter::util::makeString(const smart_ptr<unicode::UniString>& data)
{
    return new interpreter::primitive::String();
}

IObject* interpreter::util::makeCharactor(const smart_ptr<unicode::UniChar>& data)
{
    return new interpreter::primitive::Charactor();
}

IObject* interpreter::util::makeNumber(const smart_ptr<unicode::UniString>& real,
                                       const smart_ptr<unicode::UniString>& imagin,
                                       bool exact, int radix)
{
    return new interpreter::primitive::Number(real, imagin, exact, radix);
}

IObject* interpreter::util::makeBoolean(bool b)
{
    return new interpreter::primitive::Boolean(b);
}

IObject* interpreter::util::makeSymbol(const smart_ptr<unicode::UniString>& data)
{
    return new interpreter::primitive::Symbol(data);
}

IObject* interpreter::util::makeByteVector(const std::vector<data::Object>& vec)
{
    return new interpreter::primitive::ByteVector(vec);
}

IObject* interpreter::util::makeVector(const std::vector<data::Object>& vec)
{
    return new interpreter::primitive::Vector(vec);
}

void interpreter::util::concat(data::Object& c, const data::Object& dat)
{
    // Consオブジェクトであるcの末尾に、datを追加する。
    // datはconsである必要は無い。
    // cがconsでは無い場合、これは成功しない。
    assert(data::instanceof(c, type::Type(primitive::PrimitiveClass<primitive::Cons>().class_name())) &&
           "not cons");

    primitive::Cons* m = data::DataCastor<primitive::Cons>()(c);
    if (m->cdr().getSpace().isNull())
    {
        // 最初期でこうなっていたら、これにはそのまま突っ込むことができる。
        m->setCdr(dat);
    }
    else
    {
        // Consか、NullObjectに当たるまでループで回す。
        while (!m->cdr().getSpace().isNull())
        {
            if (data::instanceof(
                    m->cdr(),
                    type::Type(primitive::PrimitiveClass<primitive::Cons>().class_name())))
            {
                // consで、かつcdrが空かどうかを調べる。
                m = data::DataCastor<primitive::Cons>()(m->cdr());
                continue;
            }
            else
            {
                // Cons以外のオブジェクトがcdrに入っていた場合、これはリストでは
                // 無いため、何も行われない。
                return;
            }
        }

        // 抜けた時点で末尾であるので、cdrを設定する。
        m->setCdr(dat);
    }
}

data::Object& interpreter::util::getLastCons(data::Object& c)
{
    // 渡されたオブジェクトがconsの場合、末尾のConsを返す。
    // 末尾と判断するポイントは、cdrがnilであるか、cdrがcons以外の場合。
    // cdrがnilの時点で、リストは終了しているとみなされるため。

    assert(data::instanceof(c, type::Type(primitive::PrimitiveClass<primitive::Cons>().class_name())) &&
           "not cons");

    primitive::Cons* cons = data::DataCastor<primitive::Cons>()(c);
    smart_ptr<interpreter::IClass> cclass = cons->getClass();

    while (data::instanceof(cons->cdr(), type::Type(cclass->class_name())))
    {
        // consである間、後ろに移動しつづけてみる。
        cons = data::DataCastor<primitive::Cons>()(cons->cdr());
    }

    return cons->cdr();
}
