#include <exception>

#include "reader.h"
#include "transcoder.h"

using namespace utakata;
using namespace utakata::reader;

ReaderException::ReaderException(std::string str) : str_(str)
{}

const char* ReaderException::what() const throw()
{
    return str_.c_str();
}

//======================================================================

StreamReader::StreamReader(smart_ptr<std::istream>& s,
                           const smart_ptr<transcoder::ITranscoder>& t) :
    stream_(s), trans_(t), pos_(0)
{
}

unsigned long StreamReader::read()
{
    // 現在利用保持しているstreamから1バイト分だけ読出す。
    // eofの場合、std::istream::traits_type::eofを返す。

    validate();
    
    if (stream_->eof() || !stream_->good())
    {
        return std::istream::traits_type::eof();
    }
    ++pos_;

    return trans_->encode(*stream_);
}

std::vector<unsigned long> StreamReader::read(size_t num)
{
    // 指定したサイズ分のbyteを読出す。
    // 指定サイズが読み出しきれず、eofになった場合には、eofは返さない。
    validate();
    
    if (num <= 0)
    {
        throw ReaderException("must be greater than zero of read size");
    }

    if (stream_->eof() || !stream_->good())
    {
        return std::vector<unsigned long>();
    }

    std::vector<unsigned long> ret;
    for (unsigned long i = 0; i < num; ++i)
    {
        unsigned long tmp = trans_->encode(*stream_);
        ret.push_back(tmp);
    }

    pos_ += num;

    return ret;
}

unsigned long StreamReader::peek()
{
    // 一度読みだして、読み出した分だけい再度設定する。
    validate();
    
    if (stream_->eof() || !stream_->good())
    {
        return std::istream::traits_type::eof();
    }

    size_t prev = stream_->tellg();
    unsigned long r = trans_->encode(*stream_);
    for (size_t i = stream_->tellg(); i > prev; --i)
    {
        stream_->unget();
    }

    return r;
}

void StreamReader::unget(unsigned long ch)
{
    // 渡したchを元に戻す。成功しない場合には例外が発生する。
    validate();
    std::vector<int> dec = trans_->decode(ch);

    std::vector<int>::reverse_iterator beg = dec.rbegin(), e = dec.rend();
    for (; beg != e; ++beg)
    {
        stream_->putback(*beg);
    }

    --pos_;
}

size_t StreamReader::pos() const
{
    return pos_;
}

bool StreamReader::isEOF()
{
    // 現在streamが終端かどうかを返す。
    validate();
    return stream_->eof();
}

void StreamReader::validate()
{
    if (stream_.isNull())
    {
        throw ReaderException("stream must valid instantce but NULL!");
    }
}
