#ifndef _SIMPLETEST_H_
#define _SIMPLETEST_H_

#include <vector>
#include <string>
#include <exception>
#include <iostream>
#include <functional>

#include "smart_function_callback.h"

namespace simpletest {

//     C++にて、シンプルなテストを実行するための簡易環境。
//     基本的な使いかたは次のようになる。

//     #include "simpletest.h"
//     #include <assert.h>

//     bool hoge() {
//         int o = 100;
//         if (o != 100)
//         {
//             return false;
//         } else {
//             return true;
//         }
//     }

//     int main(int argc, char *argv[])
//     {
//         simpletest::SimpleTestSuite suite;
//         suite.addTester(fc::screate(hoge));
//         suite.run();

//         return 0;
//     }

//     出力結果 ->
//     ============================================================
//     - running...
//     ============================================================
//     - SimpleTestSuite Results
//     - ran tests 1 --> success [1] : failed [0] : exception [0]
//     ============================================================
    
//     という形で、非常にシンプルな形で使用することができる。
//     基本的に、add_suiteで追加できるのはbool (*T)()な関数のみである。
//     そして、追加した関数がtrueを返せば成功、falseを返せば失敗として、
//     一応例外が実行されたかどうかをチェックし、例外が投げられたことで関数が
//     終了している場合、それも記述される。

    class SimpleTestRunner;

    class SimpleTestAsserter;
    
    class SimpleTestSuite
    {
        // セパレータ以外の文字列の表記時に利用されるprefix
        const std::string name_;
        const std::string prefix_;

    public:
        
        SimpleTestSuite(const std::string& name, bool enableCatch = true);
        SimpleTestSuite(const std::string& name, std::ostream& strm,
                        bool enableCatch = true);
        virtual ~SimpleTestSuite(){}

        // 指定された形式の関数コールバックを渡す。
        // 返り値以外は指定されてりない。
        bool addTester(const sfcr::sfc_r<bool>::type& func);

        // ostreamが指定された場合、なんらかの媒体に出力することができ
        // る。
        // デフォルトで指定されない場合には、std::coutが利用される。
        int run(bool enableCatch = true);

        // assertを行う部分を切り出したクラスを返す。
        smart_ptr<SimpleTestAsserter> getAsserter();

    private:

        // テストの開始と終了を表示するための関数
        void printBeginMessage();
        void printEndMessage();

    private:

        std::vector<sfcr::sfc_r<bool>::type> testers_;
        std::ostream& strm_;

        smart_ptr<SimpleTestRunner> runner_;

    };

    
    //================================================================================

    class SimpleTestRunner
    {
        /**
           テスト結果をシンプルに表示するだけの機能を持つ。
        */
    public:
        SimpleTestRunner(std::ostream& strm,
                         const std::string& prefix,
                         bool enableCatch);
        
        void printSuccess();
        void printFail();
        void printException(const std::exception& e);

        // 自身をそのまま実行可能な状態とする。
        template<class T>
        void operator()(T& test)
            {
                // デフォルトで有効となっているenableCatchが
                // 無効である場合、例外が一番上まで伝播する。
                if (pImpl->enable)
                {
                    try {
                        test->run();
                        innerCheck(test->getResult());
                    } catch (std::exception& e) {
                        innerCheck(e);
                    }
                }
                else
                {
                    test->run();
                    innerCheck(test->getResult());
                }
            }

        int getSuccess() const {return pImpl->success;}
        int getFail() const {return pImpl->fail;}
        int getException() const {return pImpl->exception;}

    private:

        void innerCheck(bool b);
        void innerCheck(std::exception& e);

    private:

        std::ostream& strm_;
        const std::string prefix_;

        struct PImpl;
        smart_ptr<PImpl> pImpl;

        struct PImpl
        {
            PImpl() : success(0), fail(0), exception(0), count(0), enable(false) {}
            int success;
            int fail;
            int exception;
            int count;
            bool enable;
        };

    };

    //================================================================================
    // 単純なassertを行うためだけのクラス。
    //================================================================================

    template<class T, class S>
    struct equal : public std::binary_function<T, S, bool>
    {
        // 渡された二項が等しいことを表すためのテンプレート。
        // 基本的には==のオペレータオーバーロードが利用されるはず
        // なので特に不要とも思われるけれど、他のオペレータを
        // 用いることができるようにもしておきたいため。
        bool operator()(const T& f1, const S& f2) {
            return (f1 == f2);
        }

    };

    template<class T, class S>
    struct not_equal : public std::binary_function<T, S, bool>
    {
        // 渡された二項が等しくないことを表すテンプレート。
        // 基本的には!=のオペレータオーバーロードが利用されるはず
        // なので特に不要とも思われるけれど、他のオペレータを
        // 用いることができるようにもしておきたいため。
        bool operator()(const T& f1, const S& f2) {
            return (f1 != f2);
        }
    };

    class SimpleTestAsserter
    {
    public:
        SimpleTestAsserter(std::ostream& strm,
                            const std::string& prefix);
        SimpleTestAsserter(const SimpleTestAsserter& asserter);
        virtual ~SimpleTestAsserter() {}

        SimpleTestAsserter& operator=(const SimpleTestAsserter& asserter);

        // 外部から利用される簡単なassert。
        // 基本的にはstd::equal_toを利用することになっている。
        // それ以外も同様に定義されているが、独自にbinary_functionから派生して
        // 作成することも当然許可されている。
        // binary_functionから派生する場合、返り値はboolにすること。
        template<class First, class Second, class Compare>
        void check_(const First& f, const Second& s, const std::string& mes, Compare c) {
            if (!c(f, s))
            {
                // 失敗したら簡単なメッセージを設定することにする。
                if (!mes.empty())
                {
                    strm_ << prefix_ << "fail assert between [" << f << "] and [" << s << "] "
                          << std::endl;
                    strm_ << (prefix_ + prefix_) << "message is --> " << mes << std::endl;
                }
                else
                {
                    strm_ << prefix_ << "fail assert between [" << f << "] and [" << s << "]" << std::endl;
                }
                ++count_;
            }
        }

        template<class First, class Second, class Compare>
        void checkOnly_(const First& f, const Second& s, const std::string& mes, Compare c) {
            if (!c(f, s))
            {
                // 値を表示せず、成功したか失敗したかだけを表示する。
                if (!mes.empty())
                {
                    strm_ << prefix_ << "fail assert number : " << count_ + 1 << std::endl;
                    strm_ << (prefix_ + prefix_) << "message is --> " << mes << std::endl;
                }
                else
                {
                    strm_ << prefix_ << "fail assert number : " << count_ + 1 << std::endl;
                }
                ++count_;
            }
        }

        // デフォルトでstd::equal_toを比較関数として利用する。
        // 比較関数を渡す必要のない場合にはこれが利用される。
        template<class First, class Second>
        void check(const First& f, const Second& s, const std::string& mes = std::string()) {
            check_(f, s, mes, equal<First,Second>());
        }

        template<class First, class Second>
        void checkOnly(const First& f, const Second& s, const std::string& mes = std::string()) {
            checkOnly_(f, s, mes, equal<First, Second>());
        }

        // assertの結果が満足いくものかどうかを返す。
        bool isOk() const;

    private:

        std::ostream& strm_;
        std::string prefix_;

        int count_;

    };
    

};

#endif /* _SIMPLETEST_H_ */
