#include <iostream>
#include <assert.h>
#include <sstream>

#include "unicode.h"
#include "reader.h"
#include "sublexer_impl.h"
#include "lexeme_impl.h"
#include "delimiter.h"
#include "number_lexer.h"
#include "lexeme_id.h"

using namespace utakata;
using namespace utakata::unicode;

smart_ptr<lexeme::ILexeme> sublexer::FirstLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                     smart_ptr<sublexer::ISubLexer>& next)
{
    // chにはlexerから渡された、今回読みだした文字が渡されている。

    UniString str;

    // 最初に実行されるので、最初の空白を読み飛ばす。
    
    UniChar ch(stream->read());
    {
        lexer_delimiter::Whitespace sps;
        while (sps(ch)) {
            ch = stream->read();
        }
    }
    // ここまで来た段階で、chには調べる対象となる文字が入っている。

    if (stream->isEOF())
    {
        return lexeme::makeEOS();
    }

    smart_ptr<lexeme::ILexeme> ret;
    // ()[]`'が最初だった場合、それぞれを返してやる。
    
    if (ch == '(' || ch == '[')
    {
        ret = lexeme::makeOpenParen();
    }
    else if (ch == ')' || ch == ']')
    {
        ret = lexeme::makeCloseParen();;
    }
    else if (ch == '`')
    {
        ret = lexeme::makeBackQuote();
    }
    else if (ch == '\'')
    {
        ret = lexeme::makeQuote();
    }
    else if (ch == '.')
    {
        // 次の文字がデリミタで終了していなければならない。
        if (lexer_delimiter::Normal()(UniChar(stream->peek())))
        {
            ret = lexeme::makeDot();
        }
    }
    else if (unicode::is_numeric(ch))
    {
        // 先頭が数値だった場合、これは10進数だと判断して次に進む。
        next.add(new sublexer::NumberLexer(UniString(ch, 1)));
        ret = smart_ptr<lexeme::ILexeme>();
    }
    else if (unicode::is_alpha(ch) ||
             ch == '!' || ch == '$' ||
             ch == '%' || ch == '&' ||
             ch == '*' || ch == '/' ||
             ch == ':' || ch == '<' ||
             ch == '>' || ch == '?' ||
             ch == '^' || ch == '_' ||
             ch == '~' || ch == '+' ||
             ch == '.')
    {
        // やたらと多いが、上記のどれかである場合には、identityとして解析を
        // 開始させる。
        next.add(new sublexer::IdentifierLexer(UniString(ch, 1)));
        ret = smart_ptr<lexeme::ILexeme>();
    }
    else if (ch == ';')
    {
        // 一行コメントになるため、そちらに任せる。
        next.add(new sublexer::OneLineCommentLexer());
    }
    else if (ch == '"')
    {
        // 先頭が"の場合、stringと判断される。
        next.add(new sublexer::StringLexer());
    }

    if (!ret.isNull())
    {
        return ret;
    }
    else if (next.isNull())
    {
        // ここ以降になると、決定するまで読んでから決定させる。
        str += ch;
    
        ret = lex_(str, stream, next);
        if (!ret.isNull())
        {
            return ret;
        }
    }

    // ここに来る場合、次の解析結果が存在する場合のみである。
    if (next.isNull() && ret.isNull())
    {
        throw sublexer::LexException(stream->pos(), "lex exception");
    }
    return smart_ptr<lexeme::ILexeme>();
}

smart_ptr<lexeme::ILexeme> sublexer::FirstLexer::lex_(const utakata::unicode::UniString& str,
                                                      smart_ptr<utakata::reader::StreamReader> stream,
                                                      smart_ptr<ISubLexer>& next)
{
    if (str[0] == ',')
    {
        UniChar ch = stream->peek();
        // 次の一文字で決定できる。
        if (ch == '@')
        {
            stream->read();
            return lexeme::makeUnquoteSplicing();
        }
        else
        {
            // 次の文字が@では無い場合、とりあえず,で返す。
            return lexeme::makeUnquote();
        }
    }
    else if (str[0] == '#')
    {
        // コメントなどに繋がる時もあるため、面倒。
        UniChar ch = stream->peek();
        if (ch == '\'')
        {
            stream->read();
            return lexeme::makeSyntax();
        }
        else if (ch == '`')
        {
            stream->read();
            return lexeme::makeQuasiSyntax();
        }
        else if (ch == ',')
        {
            // 更に次の文字を読む。
            stream->read();
            ch = stream->peek();
            if (ch == '@')
            {
                stream->read();
                return lexeme::makeUnsyntaxSplicing();
            }
            else
            {
                return lexeme::makeUnsyntax();
            }
        }
        else if (ch == '|')
        {
            // ネストコメントの開始という判断がなされ、ネストコメント
            // 用の字句解析に移る。
            stream->read();
            next.add(new sublexer::NestedCommentLexer());
        }
        else if (ch == 'e' || ch == 'E' ||
                 ch == 'i' || ch == 'I' ||
                 ch == 'b' || ch == 'B' ||
                 ch == 'o' || ch == 'O' ||
                 ch == 'd' || ch == 'D' ||
                 ch == 'x' || ch == 'X')
        {
            // 数値のプレフィックス、もしくはexactnessになるため、これは
            // 数値として判定させる。
            stream->read();
            next.add(new sublexer::NumberLexer(str + ch));
        }
        else if (ch == '\\')
        {
            // #\の場合、次には文字が出てくるはずなので、そのまま進む。
            stream->read();
            next.add(new sublexer::CharactorLexer());
        }
        else if (ch == 't' || ch == 'f' ||
                 ch == 'T' || ch == 'F')
        {
            stream->read();
            next.add(new sublexer::BooleanLexer(ch));
        }
        else if (ch == 'v')
        {
            // この場合、bytevectorとなる可能性がある。
            stream->read();
            next.add(new sublexer::ByteVectorLexer(ch));
        }
        else if (ch == '(')
        {
            stream->read();
            return lexeme::makeVector();
        }
    }
    else if (str[0] == '-')
    {
        // 先頭が-の場合、次の文字を見てから決める。
        // 基本的には数値だが、次の文字によってはidentifierになりうる。
        UniChar ch(stream->peek());
        lexer_delimiter::Normal nor;
        if (ch == '>' || nor(ch))
        {
            next.add(new sublexer::IdentifierLexer(str));
        }
        else
        {
            next.add(new sublexer::NumberLexer(str));
        }
    }

    return smart_ptr<lexeme::ILexeme>();
}


//================================================================================

smart_ptr<lexeme::ILexeme> sublexer::StringLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                      smart_ptr<sublexer::ISubLexer>& next)
{
    // 文字列を解析する。

    lexer_delimiter::Normal nor;
    lexer_delimiter::String st;
    lexer_delimiter::Whitespace w;
    unicode::UniString str;

    while (!stream->isEOF()) {
        UniChar tmp(stream->peek());
        if (st(tmp)) {
            // readしておく。
            stream->read();
            break;
        }

        if (tmp == '\\')
        {
            // 空白及び改行を飛ばす。
            while (!stream->isEOF() &&
                   (w(tmp) || nor(tmp)))
            {
                tmp = stream->read();
            }
        } else {
            str += stream->read();
        }
    }

    return lexeme::makeString(str);
}

//================================================================================

smart_ptr<lexeme::ILexeme> sublexer::NestedCommentLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                             smart_ptr<sublexer::ISubLexer>& next)
{
    // ネストしたコメントを解釈する。基本的には改行も関係無く処理する。
    int count = 1;
    while (!stream->isEOF() && count > 0) {
        unicode::UniChar tmp(stream->read());

        if (tmp == '#')
        {
            // 次の文字を調べる。
            unicode::UniChar t2(stream->read());
            if (t2 == '|')
            {
                ++count;
            }
        }
        else if (tmp == '|')
        {
            // 次の文字を調べる。
            unicode::UniChar t2(stream->read());
            if (t2 == '#')
            {
                --count;
            }
        }
    }

    return smart_ptr<lexeme::ILexeme>();
}

//================================================================================

sublexer::NumberLexer::NumberLexer(const unicode::UniString& str) :
    BINARY(2), OCTET(8), DECIMAL(10), HEX(16),
    str_(new unicode::UniString(str)), exact_(false), prefix_(0)
{
}

smart_ptr<lexeme::ILexeme> sublexer::NumberLexer::innerLex_(smart_ptr<reader::StreamReader> stream,
                                                            smart_ptr<sublexer::ISubLexer>& next,
                                                            const unicode::UniString& str)
{

    // prefixをチェックして、返すべきものを決定する。
    if (prefix_ == BINARY)
    {
        return number::Number<2>().lex(stream, next, str, exact_);
    }
    else if (prefix_ == OCTET)
    {
        return number::Number<8>().lex(stream, next, str, exact_);
    }
    else if (prefix_ == DECIMAL)
    {
        return number::Number<10>().lex(stream, next, str, exact_);
    }
    else if (prefix_ == HEX)
    {
        return number::Number<16>().lex(stream, next, str, exact_);
    }
    else
    {
        return smart_ptr<lexeme::ILexeme>();
    }
}

void sublexer::NumberLexer::checkExactness_(const unicode::UniString& str)
{
    // 正確性をチェックする。2文字目だけを調べればそれですむ。
    if (str[1] == 'e' || str[1] == 'E')
    {
        exact_ = true;
    }
    else if (str[1] == 'i' || str [1] == 'I')
    {
        exact_ = false;
    }
}

unsigned char sublexer::NumberLexer::getPrefix_(const unicode::UniString& str)
{
    // 2文字目の値を見て判断を返す。
    if (str[1] == 'b' ||
        str[1] == 'B')
    {
        return BINARY;
    }
    else if (str[1] == 'o' ||
             str[1] == 'O')
    {
        return OCTET;
    }
    else if (str[1] == 'd' ||
             str[1] == 'D')
    {
        return DECIMAL;
    }
    else if (str[1] == 'x' ||
             str[1] == 'X')
    {
        return HEX;
    }

    return DECIMAL;
}

smart_ptr<lexeme::ILexeme> sublexer::NumberLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                      smart_ptr<sublexer::ISubLexer>& next)
{
    // 数値を解釈する。
    lexer_delimiter::Exactness e;
    lexer_delimiter::Prefix p;
    

    // 先頭はすでに読みとばされている。次に続く可能性があるため、継続して調査をしてみる。

    if (p(*str_))
    {
        prefix_ = getPrefix_(*str_);
        // 次にexactnessが続いているかどうかを調べる。
        unicode::UniString s = unicode::substring(*str_, 2);
        if (e(s))
        {
            // Exactnessである場合、実際にそうであるかどうかをチェックする。
            checkExactness_(s);
        }
        s = unicode::substring(s, 2);
        return innerLex_(stream, next, s);
    }
    else if (e(*str_))
    {
        // Exactnessであるため、次にprefixの必要がある。
        checkExactness_(*str_);

        unicode::UniString s = unicode::substring(*str_, 2);
        if (p(s))
        {
            prefix_ = getPrefix_(s);
        }
        else
        {
            prefix_ = DECIMAL;
        }
        return innerLex_(stream, next, s);
    }
    else
    {
        // どっちでもない場合、基本的に10進数として処理する。
        prefix_ = DECIMAL;
        return innerLex_(stream, next, *str_);
    }
}

//================================================================================

sublexer::IdentifierLexer::IdentifierLexer(const unicode::UniString& str) :
    str_(new unicode::UniString(str))
{}

smart_ptr<lexeme::ILexeme> sublexer::IdentifierLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                          smart_ptr<sublexer::ISubLexer>& next)
{
    // Identityを解析する。解釈自体は結構簡単。
    
    // 一文字目はすでに判定されているため、二文字目以降で判定する。
    // 二文字名以降でなければ判定できないものもあるため、一度まとめてみる。
    lexer_delimiter::Normal nor;
    unicode::UniString& str = *str_;
    if (str[0] == '+')
    {
        return lexeme::makeIdentifier(str);
    }
    else if (str[0] == '-')
    {
        // 次の文字がデリミタの場合には、そのまま返す。
        UniChar next(stream->peek());
        if (nor(next))
        {
            return lexeme::makeIdentifier(str);
        }
        else if (next == '>')
        {
            // この場合、そのまま続けてもよし。
        }
    }

    while (!stream->isEOF()) {
        UniChar tmp(stream->peek());
        if (nor(tmp)) {
            break;
        }
        str += stream->read();
    }
    
    return lexeme::makeIdentifier(str);
}

//================================================================================

smart_ptr<lexeme::ILexeme> sublexer::CharactorLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                         smart_ptr<sublexer::ISubLexer>& next)
{
    // #\に続く文字名、あるいは#\xに続く16進数のチェックを行なう。
    // ここに来た時点で、#\までは読まれている。
    UniChar ch(stream->read());
    bool flag = ch == 'x' ? true : false;

    // Hexを解釈する。解釈部分を書くと面倒になるため、単純な別関数にしておく。
    UniString str(ch, 1);
    lexer_delimiter::Normal nor;
    lexer_delimiter::HexValue hex;
    while (!stream->isEOF()) {
        UniChar tmp(stream->peek());
        if (nor(tmp))
        {
            break;
        }
        else if (flag && !hex(tmp))
        {
            // hex valueではなかった場合には、これはエラーであると
            // して返す。
            std::stringstream ss;
            ss << tmp.getRawCode() << "は16進数中の文字名として利用できません";
            throw sublexer::LexException(stream->pos(), ss.str());
        }
        str += stream->read();
    }

    // 抜けた時点で、構成文字は問題ないこととなっている。ここからは、対象とする一文字を
    // 生成する処理を開始する。
    UniChar c(parseCharSpec(str));

    return lexeme::makeCharactor(c);
}

unsigned long sublexer::CharactorLexer::parseCharSpec(const UniString& c)
{
    return 0x0;
}

//================================================================================

sublexer::BooleanLexer::BooleanLexer(const unicode::UniChar& ch) : ch_(new unicode::UniChar(ch))
{
}

smart_ptr<lexeme::ILexeme> sublexer::BooleanLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                         smart_ptr<sublexer::ISubLexer>& next)
{
    // true/falseを判別する必要は特に無いため、次の文字が区切り文字だかどうか
    // だけを気にする。

    if (!stream->isEOF())
    {
        UniChar ch(stream->peek());

        lexer_delimiter::Normal nor;
        if (!nor(ch))
        {
            // hex valueではなかった場合には、これはエラーであると
            // して返す。
            std::stringstream ss;
            ss << ch.getRawCode() << "はbooleanの直後では認識されません";
            throw sublexer::LexException(stream->pos(), ss.str());
        }
    }

    return lexeme::makeBoolean(ch_);
}

//================================================================================

sublexer::ByteVectorLexer::ByteVectorLexer(const unicode::UniChar& ch) : ch_(new unicode::UniChar(ch))
{
}

smart_ptr<lexeme::ILexeme> sublexer::ByteVectorLexer::lex(smart_ptr<reader::StreamReader> stream,
                                                         smart_ptr<sublexer::ISubLexer>& next)
{
    // この時点で#vまでは読まれているため、以降2文字がvu(であるかどうかを
    // 調べる。

    UniString str(utakata::unicode::convert(stream->read(3)));
    if (str != utakata::unicode::convert("u8("))
    {
        std::stringstream ss;
        ss << "バイトベクタとして認識できません";
        throw sublexer::LexException(stream->pos(), ss.str());
    }

    return lexeme::makeByteVector();
}


//================================================================================


smart_ptr<lexeme::ILexeme> sublexer::OneLineCommentLexer::lex(
    smart_ptr<reader::StreamReader> stream,
    smart_ptr<sublexer::ISubLexer>& next)
{


    // 一行コメントであるかどうかを返す。
    // なおコメントは原則として全て読み飛ばされるため、解釈が完了した
    // 後、FirstLexerが再び返される。
    lexer_delimiter::LineEnding end;
    while (!stream->isEOF()) {
        // 改行が出力されるまで全て読み飛ばす。
        UniChar ch = stream->read();
        if (end(ch, stream))
        {
            // この時点で、読み飛ばしまで行われている。
            next.add(new sublexer::FirstLexer());
            break;
        }
    }
    return smart_ptr<lexeme::ILexeme>();
}


//================================================================================

sublexer::LexException::LexException(size_t pos, std::string str) : pos_(pos),
                                                                    str_()
{
    // エラーメッセージを定義する。
    std::stringstream ss;
    ss << "lex error ! -- pos : [" << pos << "] message : [" << str << "]" << std::endl;
    str_ = ss.str();
}

const char* sublexer::LexException::what() const throw()
{
    return str_.c_str();
}
