#include <iostream>
#include <sstream>
#include <string>
#include <functional>

#include "../simpletest.h"

#include "../textarrayformat.h"
#include "../unicode.h"
#include "../parser.h"
#include "../lexeme_id.h"
#include "../lexeme.h"

#include "../lexer.h"

#include "../tree.h"
#include "../utf8_transcoder.h"
#include "../reader.h"
#include "../object.h"
#include "../gc.h"
#include "../type.h"
#include "../primitive.h"
#include "../data_castor.h"
#include "../data_structure.h"
#include "../primitive_class.h"

using namespace std;
using namespace utakata;

data::Object cadr(interpreter::primitive::Cons* c)
{
    return data::DataCastor<interpreter::primitive::Cons>()(c->cdr())->car();
}

data::Object cdar(interpreter::primitive::Cons* c)
{
    return data::DataCastor<interpreter::primitive::Cons>()(c->car())->cdr();
}

interpreter::primitive::Cons* tocons(data::Object o)
{
    return data::DataCastor<interpreter::primitive::Cons>()(o);
}


bool parser_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << "(+ 12 24)" << endl;
    ss << "==========" << endl;
    ss << "(+ (- 12 \"hoge\") 24)" << endl;
    ss << "==========" << endl;
    ss << "((- 12 \"hoge\") '24)" << endl;
    ss << "==========" << endl;
    ss << "(define hoge (lambda () (- 12 24)))" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get(2)));
    smart_ptr<utakata::reader::StreamReader> st(
        new utakata::reader::StreamReader(formats,
                                          smart_ptr<transcoder::ITranscoder>(
                                              new transcoder::UTF8Transcoder())));

    std::vector<smart_ptr<lexeme::ILexeme> > lexemes;

    smart_ptr<utakata::lexer::Lexer> lexer(new lexer::Lexer());

    parser::Parser p(lexer);
    // parserから生成されるオブジェクトは完全なdungring objectとなる。
    // primitiveなデータに関しては、singletonなTypeHashを保持する、
    // primitive factoryが利用される。
    gc::GarbageCollector gc;
    smart_ptr<data::Object> t = p.parse(st, gc);

    // 先頭はconsであるはず。
    interpreter::primitive::PrimitiveClass<interpreter::primitive::Cons> consclass;
    asserter->check(data::instanceof(*t, type::Type(
                                         consclass.class_name())), true,
                    "cons error1");

    // 更に内部についても、同様にConsであるはず。((の部分。
    data::Object o = data::DataCastor<interpreter::primitive::Cons>()(*t)->car();
    asserter->check(data::instanceof(o, type::Type(consclass.class_name())), true,
                    "cons error 2");

    // さらにcarを取得してみる。ここでのcarはつまり `-` なのでシンボル
    {
        using namespace utakata::interpreter::primitive;
        data::Object o2 = data::DataCastor<interpreter::primitive::Cons>()(o)->car();
        asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<Symbol>().class_name())), true,
                        "symbol error1");
    }

    // cdrはリストなので、当然だがcons。
    o = data::DataCastor<interpreter::primitive::Cons>()(o)->cdr();
    asserter->check(data::instanceof(o, type::Type(consclass.class_name())), true,
                    "cons error 2");

    {
        // 数値が来ている。
        using namespace utakata::interpreter::primitive;
        
        data::Object o2 = data::DataCastor<interpreter::primitive::Cons>()(o)->car();
        asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<Number>().class_name())), true,
                        "number error");
    }

    o = data::DataCastor<interpreter::primitive::Cons>()(o)->cdr();
    {
        // 文字列が来ている。
        using namespace utakata::interpreter::primitive;
        
        data::Object o2 = data::DataCastor<interpreter::primitive::Cons>()(o)->car();
        asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<String>().class_name())), true,
                        "string error");
    }

    // ここでcdrを取得すると、'24に繋がる。
    o = data::DataCastor<interpreter::primitive::Cons>()(*t)->cdr();
    o = data::DataCastor<interpreter::primitive::Cons>()(o)->car();
    data::Object o2 = data::DataCastor<interpreter::primitive::Cons>()(o)->car();

    
    // ここでquoteである。
    asserter->check(data::instanceof(o2, type::Type(interpreter::primitive::PrimitiveClass<interpreter::primitive::Quote>().class_name())), true,
                    "quote Error");

    // cdrはCons
    o2 = data::DataCastor<interpreter::primitive::Cons>()(o)->cdr();
    {
        // 数値が来ている。
        using namespace utakata::interpreter::primitive;
        
        data::Object o3 = data::DataCastor<interpreter::primitive::Cons>()(o2)->car();
        asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<Number>().class_name())), true,
                        "number error");
    }

    data::Object o3 = data::DataCastor<interpreter::primitive::Cons>()(o2)->cdr();
    asserter->check(data::instanceof(o3, type::Type(interpreter::primitive::PrimitiveClass<
                                                    interpreter::primitive::Nil>().class_name())), true,
                    "nil Error");
    
    return asserter->isOk();
}

bool parser_test2(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << "(+ 12 24)" << endl;
    ss << "==========" << endl;
    ss << "(+ (- 12 \"hoge\") (14 . 1))" << endl;
    ss << "==========" << endl;
    ss << "((- 12 \"hoge\") '24)" << endl;
    ss << "==========" << endl;
    ss << "(define hoge (lambda () (- 12 24)))" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get(1)));
    smart_ptr<utakata::reader::StreamReader> st(
        new utakata::reader::StreamReader(formats,
                                          smart_ptr<transcoder::ITranscoder>(
                                              new transcoder::UTF8Transcoder())));

    std::vector<smart_ptr<lexeme::ILexeme> > lexemes;

    smart_ptr<utakata::lexer::Lexer> lexer(new lexer::Lexer());

    parser::Parser p(lexer);
    // parserから生成されるオブジェクトは完全なdungring objectとなる。
    // primitiveなデータに関しては、singletonなTypeHashを保持する、
    // primitive factoryが利用される。
    gc::GarbageCollector gc;
    smart_ptr<data::Object> t = p.parse(st, gc);

    using namespace utakata::interpreter::primitive;

    data::Object o = tocons(*t)->car();
    // ここはシンボル
    asserter->check(data::instanceof(o, type::Type(PrimitiveClass<Symbol>().class_name())),
        true);

    o = tocons(*t)->cdr();
    asserter->check(data::instanceof(o, type::Type(PrimitiveClass<Cons>().class_name())),
        true);

    // (- 12 "hoge")
    data::Object o2 = tocons(o)->car();
    asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<Cons>().class_name())),
        true);

    data::Object o3 = tocons(o2)->car();
    asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<Symbol>().class_name())),
        true);

    // 12
    o3 = cadr(tocons(o2));
    asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<Number>().class_name())),
        true);

    // "hoge"
    o3 = cadr(tocons(tocons(o2)->cdr()));
    asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<String>().class_name())),
        true);

    // ((14 . 1))
    o2 = tocons(o)->cdr();

    // (14 . 1)
    o2 = tocons(o2)->car();

    // 14
    o3 = tocons(o2)->car();
    asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<Number>().class_name())),
                    true, "dot first");

    // 1
    o3 = tocons(o2)->cdr();
    asserter->check(data::instanceof(o3, type::Type(PrimitiveClass<Number>().class_name())),
                    true, "dot second");

    return asserter->isOk();
}

bool parser_test3(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    std::stringstream ss;
    ss << "==========" << endl;
    ss << "(+ 12 24)" << endl;
    ss << "==========" << endl;
    ss << "(+ (- 12 \"hoge\") (14 . 1))" << endl;
    ss << "==========" << endl;
    ss << "((- 12 \"hoge\") '24)" << endl;
    ss << "==========" << endl;
    ss << "(define hoge (lambda () (- 12 24)))" << endl;

    textarrayformat::TextArrayReader reader(ss);
    smart_ptr<istream> formats(new stringstream(reader.get(3)));
    smart_ptr<utakata::reader::StreamReader> st(
            new utakata::reader::StreamReader(formats,
                                              smart_ptr<transcoder::ITranscoder>(
                                                      new transcoder::UTF8Transcoder())));

    std::vector<smart_ptr<lexeme::ILexeme> > lexemes;

    smart_ptr<utakata::lexer::Lexer> lexer(new lexer::Lexer());

    parser::Parser p(lexer);
    // parserから生成されるオブジェクトは完全なdungring objectとなる。
    // primitiveなデータに関しては、singletonなTypeHashを保持する、
    // primitive factoryが利用される。
    gc::GarbageCollector gc;
    smart_ptr<data::Object> t = p.parse(st, gc);

    using namespace utakata::interpreter::primitive;

    data::Object o = tocons(*t)->car();
    // define
    asserter->check(data::instanceof(o, type::Type(PrimitiveClass<Symbol>().class_name())),
                    true, "symbol define");

    // (hoge (lambda () (- 12 24)))
    o = tocons(*t)->cdr();
    data::Object o2 = tocons(o)->car();
    // hoge
    asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<Symbol>().class_name())),
                    true, "symbol hoge");

    // ((lambda () (- 12 24)))
    o = tocons(o)->cdr();
    // (lambda () (- 12 24))
    o2 = tocons(o)->car();
    asserter->check(data::instanceof(o2, type::Type(PrimitiveClass<Cons>().class_name())),
                    true, "cons");

    // lambda
    asserter->check(data::instanceof(tocons(o2)->car(),
                                     type::Type(PrimitiveClass<Cons>().class_name())),
                    true, "lambda");

    // nil
    data::Object o3 = tocons(o2)->cdr();
    asserter->check(data::instanceof(tocons(o3)->car(),
                                     type::Type(PrimitiveClass<Nil>().class_name())),
                    true, "nil");

    // ((- 12 24)) -> (- 12 24)
    o3 = cadr(tocons(o3));
    asserter->check(data::instanceof(tocons(o3)->car(),
                                     type::Type(PrimitiveClass<Symbol>().class_name())),
                    true, "-");

    // 12
    asserter->check(data::instanceof(cadr(tocons(o3)),
                                     type::Type(PrimitiveClass<Number>().class_name())),
                    true, "12");

    // (12 24)
    data::Object o4 = tocons(o3)->cdr();
    asserter->check(data::instanceof(cadr(tocons(o4)),
                                     type::Type(PrimitiveClass<Number>().class_name())),
                    true, "24");

    o4 = tocons(o4)->cdr();
    asserter->check(data::instanceof(cadr(tocons(o4)),
                                     type::Type(PrimitiveClass<Nil>().class_name())),
                    true, "nil");


    return asserter->isOk();
}

int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("parser test");
    suite.addTester(sfcr::screate(parser_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(parser_test2, suite.getAsserter()));
    // suite.addTester(sfcr::screate(parser_test3, suite.getAsserter()));
    // suite.addTester(sfcr::screate(parser_test4, suite.getAsserter()));
    suite.run();
    return 0;
}
