#include <ios>
#include <iostream>
#include <sstream>
#include <string>
#include <functional>


#include "../simpletest.h"

#include "../unicode.h"
#include "../textarrayformat.h"
#include "../reader.h"
#include "../utf8_transcoder.h"


using namespace utakata;


bool utf8_multichar_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{

    std::string tmp("あいuえお");
    // マルチバイト文字列を正しく読みだせるかどうかのチェックを
    // 含めたテスト
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    reader::StreamReader stream(ss,
                                smart_ptr<transcoder::ITranscoder>(
                                    new transcoder::UTF8Transcoder()));
    
    // 単独のreadを試す。
    unicode::UniChar ch(stream.read());
    // U+3042 = あ
    asserter->check(ch.getRawCode(), 0x3042, "raw-code isn't japanese `a`");
    // peekが問題なく働いていることのテスト
    unicode::UniChar ch2(stream.peek());
    // U+3044 = い
    asserter->check(ch2.getRawCode(), 0x3044, "raw-code isn't japanese `i`");
    unicode::UniChar ch3(stream.read());
    asserter->check(ch3.getRawCode(), 0x3044, "raw-code isn't japanese `i`");

    // asciiだとしても問題無く読みだせるはず。
    unicode::UniChar ch4(stream.peek());
    asserter->check(ch4.getRawCode(), 'u');

    // ascii文字の範囲であるかどうかを判定する関数のテスト。
    asserter->check(unicode::is_ascii(ch4), true);
    asserter->check(unicode::is_ascii(ch), false);

    // 同一コードなので比較して同じになるはず。
    asserter->check(ch == ch2, false, "chとch2が異なる");
    asserter->check(ch < ch2, true , "chよりch2が大きいはず");

    return asserter->isOk();
}

bool unicode_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // マルチバイトとascii文字の混在文字も正しく扱うことのできる
    // CUniStringのテスト
    
    std::string tmp("あいうえsssお(");

    smart_ptr<std::istream> ss(new std::stringstream(tmp));
    reader::StreamReader stream(ss,
                                smart_ptr<transcoder::ITranscoder>(
                                    new transcoder::UTF8Transcoder()));

    unicode::UniString str;
    std::vector<unsigned long> t = stream.read(5);
    std::vector<unsigned long>::iterator it = t.begin(),
        end = t.end();
    for (; it != end; ++it)
    {
        str += unicode::UniChar(*it);
    }
    
    asserter->check(str.begin()->getRawCode(), 0x3042);
    asserter->check(str.size(), 5);

    // 自分自身との加算でも問題ないことを確認する。
    str += str;
    asserter->check(str[5].getRawCode(), 0x3042);

    // 本当に基本的なもの以外は、同一の名前空間内に関数が用意されている。
    // unicodeから返されるのは、UTF8のバイト列を設定した文字列であると
    // する。
    asserter->check(unicode::substring(str,0,2)[1].getRawCode(), 0x3044);
    asserter->check(unicode::substring(str,1,2)[1].getRawCode(), 0x3046);

    return asserter->isOk();
}

bool unicode_util_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // CUniStringと共に利用するためのユーティリティ関数の操作を
    // 行う。
    
    std::string tmp("あいうえsssお");

    smart_ptr<std::istream> ss(new std::stringstream(tmp));
    reader::StreamReader stream(ss,
                                smart_ptr<transcoder::ITranscoder>(
                                    new transcoder::UTF8Transcoder()));

    unicode::UniString str(unicode::convert(stream.read(5)));
    unicode::UniString str2(unicode::convert(stream.read(3)));
    unicode::UniString str3 = str;

    // 挿入してみる。
    str.insert(str.begin(), str2.begin(), str2.end());
    asserter->check(str[0].getRawCode(), 's');

    str3.insert(str3.begin() + 1, str2.begin(), str2.begin() + 1);
    asserter->check(str3[1].getRawCode(), 's');

    return asserter->isOk();
}


int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("UTF-8 文字列テスト");
    suite.addTester(sfcr::screate(utf8_multichar_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(unicode_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(unicode_util_test, suite.getAsserter()));
    suite.run();
    return 0;
}
