#include <iostream>
#include <sstream>
#include <string>
#include <algorithm>

#include <functional>

#include "../simpletest.h"
#include "../utf8_transcoder.h"
#include "../reader.h"
#include "../textarrayformat.h"

using namespace std;
using namespace utakata;

struct identity_local
{
    char operator()(int t) {return static_cast<char>(t);}
};


bool utf8_charcheck_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // UTF-8の文字を認識するテスト。

    // ソースがすでにUTF8で記述されているので、そのまま簡単にだが
    // テストが行える。
    std::string tmp("UTF8");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    reader::StreamReader stream(ss,
                                smart_ptr<transcoder::ITranscoder>(
                                    new transcoder::UTF8Transcoder()));

    // 単独のreadを試す。
    asserter->check(stream.read(), 'U', "error compare to U!");
    // peekが問題なく働いていることのテスト
    asserter->check(stream.peek(), 'T', "error compare to T and peek"); 
    asserter->check(stream.read(), 'T', "error compare to T and read");

    // ここの時点ではまだeofではない。
    asserter->check(stream.isEOF(), false, "not end of file");
    // 読み込んだ文字数を返すメンバ関数のチェック。
    asserter->check(stream.pos(), 2, "wrong position");

    // 複数文字の読み出しのチェック
    std::vector<unsigned long> f = stream.read(2);
    std::string t("", f.size());

    std::transform(f.begin(), f.end(), t.begin(),
                   identity_local());
    
    asserter->check(t, "F8", "複数文字チェック");

    // この時点でeofがtrueとなる
    asserter->check(stream.isEOF(), false, "why not eof?");
    
    return asserter->isOk();
}

bool utf8_decode_test(smart_ptr<simpletest::SimpleTestAsserter> asserter)
{
    // Unicodeスカラー値からUTF-8のデータへとデコードする。

    // ソースがすでにUTF8で記述されているので、そのまま簡単にだが
    // テストが行える。
    std::string tmp("あいうえお");
    // 1byteの文字のみの場合
    smart_ptr<std::istream> ss(new std::stringstream(tmp));

    reader::StreamReader stream(ss,
                                smart_ptr<transcoder::ITranscoder>(
                                    new transcoder::UTF8Transcoder()));

    // 原則として、一文字ずつ分割できればそれで問題は発生しない。
    transcoder::UTF8Transcoder t;
    unsigned long p = stream.peek();
    unsigned long p2 = stream.peek();

    asserter->check(p, p2, "not equal p and p2");
    
    std::vector<char> d(3, 0);
    ss->get(&d[0], 4);
    std::vector<unsigned char> retval;
    retval.insert(retval.begin(), d.begin(), d.end());

    std::vector<int> dec = t.decode(p);
    asserter->check(dec[0] == retval[0], true, "error compare to U1!");
    asserter->check(dec[1] == retval[1], true, "error compare to U2!");
    asserter->check(dec[2] == retval[2], true, "error compare to U3!");

    return asserter->isOk();
}

int main(int argc, char *argv[])
{
    simpletest::SimpleTestSuite suite("UTF-8 基本テスト");
    suite.addTester(sfcr::screate(utf8_charcheck_test, suite.getAsserter()));
    suite.addTester(sfcr::screate(utf8_decode_test, suite.getAsserter()));
    suite.run();
    return 0;
}

