#include <iostream>

#include <assert.h>
#include <exception>

#include "tree.h"
#include <stack>

#include "unicode.h"
#include "literal_data.h"
#include "literal.h"
#include "datum_id.h"

using namespace utakata;
using namespace utakata::literal;
using namespace utakata::unicode;
using namespace utakata::syntax;

// 原則としてそれぞれ単純に呼びだされるのみ。

struct Tree::TreeNode
{
    TreeNode (node_pair p) : list(), datum (p) {}

    std::vector<smart_ptr<TreeNode> > list;

    // datumの型と値の両方を保持する。
    Tree::node_pair datum;
};

////////////////////////////
// syntax::Tree::iterator //
////////////////////////////

syntax::Tree::iterator::iterator () : current_()
{
}

syntax::Tree::iterator::iterator (const Tree::iterator& i)
{
    swap (i);
}

syntax::Tree::iterator& syntax::Tree::iterator::operator=(const Tree::iterator& rh)
{
    iterator it (rh);
    swap (it);
    return *this;
}

void syntax::Tree::iterator::swap (const Tree::iterator& t)
{
    current_ = t.current_;
}

Tree::iterator& Tree::iterator::operator++ ()
{
    // car,cdr をそれぞれ巡回する。 car,cdr それぞれに移動する場合には、
    // it を元の tree に渡して取得する。
    // 巡回の順序は次の通りとする。
    // cons -> cdr に繋がっている cons
    // ただし、 cdr に繋がっている cons が literal の場合、以降は進むことが
    // 無く、また list として判定されることはない。

    ++current_;
    return *this;
}

std::pair<DatumID, smart_ptr<literal::Literal> >& Tree::iterator::operator*()
{
    // 現在の current_が指す node のリテラルを返す。
    return (*current_)->datum;
}

std::pair<DatumID, smart_ptr<literal::Literal> >* Tree::iterator::operator->()
{
    // 現在の current_が指す node のリテラルを返す。
    return &((*current_)->datum);
}

bool Tree::iterator::operator==(const Tree::iterator& rh)
{
    // node_が指す先が同一かどうかを返す。
    // 但し、自身か相手のいずれかが nil である場合だけは例外的に、単純に DatumID のみの
    // 比較になる。
    return current_ == rh.current_;
}

//////////////////
// syntax::Tree //
//////////////////

Tree::Tree () : node_()
{
    // 初期値を設定する。
    // smart_ptr<TreeNode> p(new TreeNode(makeNilList()));
    // 最後尾には常にnilがいることにする。
}

Tree::iterator Tree::begin ()
{
    // 開始点を示すデータを返す。
    iterator it;
    it.current_ = node_.begin();
    return it;
}

Tree::iterator Tree::end ()
{
    // 終了地点を示す部分を返す。
    iterator it;
    it.current_ = node_.end();
    return it;
}

void Tree::push_back(node_pair datum)
{
    // 渡されたdatumを、現在保持しているデータ型に追加する。
    smart_ptr<TreeNode> p(new TreeNode(datum));
    node_.insert(node_.end(), p);
}

void Tree::push_back(Tree::iterator it, Tree::iterator it2,
                     node_pair datum)
{
    // itとit2までを、datumで設定されたdatumidであるcompound datum
    // に格納して追加する。
    smart_ptr<TreeNode> p(new TreeNode(datum));
    p->list.insert(p->list.end(), it.current_, it2.current_);
    node_.insert(node_.end(), p);
}

Tree::iterator Tree::in_compound_begin(Tree::iterator it)
{
    // itがcompound datumである場合に、compound datumの内部リストを返す。
    if (isCompoundDatum(it))
    {
        iterator ret;
        ret.current_ = (*it.current_)->list.begin();
        return ret;
    }
    else
    {
        // 違う場合には、endを返すことにする。
        return end();
    }
}

Tree::iterator Tree::in_compound_end(Tree::iterator it)
{
    // itがcompound datumである場合に、compound datumの内部リストを返す。
    if (isCompoundDatum(it))
    {
        iterator ret;
        ret.current_ = (*it.current_)->list.end();
        return ret;
    }
    else
    {
        // 違う場合には、endを返すことにする。
        return end();
    }
}

//////////////////////
// helper functions //
//////////////////////

bool syntax::isCompoundDatum(Tree::iterator it)
{
    if (it->first == syntax::DatumID::list ||
        it->first == syntax::DatumID::vector ||
        it->first == syntax::DatumID::byteVector)
    {
        return true;
    }
    else
    {
        return false;
    }
}

std::pair<DatumID, smart_ptr<literal::Literal> > syntax::makeCompoundDatum (
    DatumID id)
{
    // datum_id が list である node_pair を返す。
    return std::pair<DatumID, smart_ptr<literal::Literal> >(id, smart_ptr<literal::Literal>());
}

std::pair<DatumID, smart_ptr<literal::Literal> > syntax::makeNilList ()
{
    // datum_id が nil である node_pair を返す。
    return std::pair<DatumID, smart_ptr<literal::Literal> >(
        DatumID(DatumID::nil), smart_ptr<literal::Literal>());
}

std::pair<DatumID, smart_ptr<literal::Literal> > syntax::makeLexemeDatum (smart_ptr<literal::Literal> l,
    DatumID id)
{
    // datum_id が各 lexeme であり、 literal として l を持つ
    // datum を返す。
    return std::pair<DatumID, smart_ptr<literal::Literal> >(
        id, l);
}

std::pair<DatumID, smart_ptr<literal::Literal> > syntax::makeAbbreviation (smart_ptr<literal::Literal> l)
{
    // datum_id が literal であり、 literal として l を持つ
    // datum を返す。
    return std::pair<DatumID, smart_ptr<literal::Literal> >(
        DatumID (DatumID::abbreviations), l);
}

