#ifndef _TREE_H_
#define _TREE_H_

#include "smart_ptr.h"
#include <utility>              // std::pair
#include <vector>

namespace utakata {

    namespace unicode {
    
        class UniString;
    
    };

    namespace literal {
    
        struct LiteralData;
        class Literal;
    
    };

    namespace lexeme {
        class LexemeID;
    };

    namespace syntax {

        class DatumID;
        class Tree
        {
            // Scheme の構文データを表す Tree を返す。
            // 内部的には、殆ど実体そのままの cons を構成する。
            // ただし、構成している cons の内容としては、 datum でしか無い
            // ため、そのあたりが多少異なる。内部での実装は、次のように
            // なる。
            // lexeme datum -> node_pair::first = literal node_pair::second = literal::literal
            // compound datum -> node_pair::first = list, vector, bytevector node_pair::second = NULL
            // abbreviation -> node_pair::first = abbreviation, node_pair::second = literal::Literal
            // また、 lexeme datum は、 car,cdr それぞれが存在しておらず、
            // node_pair の second が設定されるのみとなる。
            // また、 iterator 及び iterator の ++ による各要素の巡回を行なうことができる。
            // iterator 中で、 nil と判断されたデータについては、その時点で
            // 巡回が停止される。
            // 空リスト≠ nil である。 nil はあくまで syntax tree の終端を示すためだけの
            // 特別な datum であり、外部にこの定義が漏れることはない。
            // 空リストは、「 car,cdr がそれぞれ isNull == true である datumid = list 」である
            // cons が対象となる。

            typedef std::pair<DatumID, smart_ptr<literal::Literal> > node_pair;
            // 実装でのみ定義される。
            struct TreeNode;

        public:

            class iterator
            {
            public:
                // ツリーを巡回し、そのノードの literal を取得することを可能とする。
                iterator ();
                iterator (const iterator& s);
                
                ~iterator () {}

                iterator& operator=(const iterator& rh);
                iterator& operator++ ();
                node_pair& operator*();
                node_pair* operator->();

                // 一致するかどうかを返す
                bool operator==(const iterator& rh);
                bool operator!=(const iterator& rh) {
                    return ! (*this == rh);
                }

                // 現在位置を表すiterator
                std::vector<smart_ptr<TreeNode> >::iterator current_;

            private:

                void swap (const iterator& s);

            };

            
        public:

            Tree ();
            virtual ~Tree () {}

            // 現在の Tree の先頭=root node と、 end = 次に指す場所が存在しない
            // node を返す。
            iterator begin ();
            iterator end ();

            // push_backによって追加を行う。
            void push_back (node_pair datum);
            void push_back (iterator b, iterator e, node_pair datum);

            // itが指すnodeの内部listの先頭を返す。
            iterator in_compound_begin(iterator it);
            iterator in_compound_end(iterator it);

            size_t size() const {return node_.size();}

        private:

            std::vector<smart_ptr<TreeNode> > node_;

        };

        // 渡した iterator が list datum であるかどうかを返す。
        bool isCompoundDatum (Tree::iterator node);

        // 渡した node が literal であるかどうかを返す。
        bool isLiteral (Tree::iterator node);

        // 各 datum を表す std::pair を作成するためのヘルパ関数。
        std::pair<DatumID, smart_ptr<literal::Literal> > makeCompoundDatum (
            DatumID id);
        
        std::pair<DatumID, smart_ptr<literal::Literal> > makeNilList ();
        std::pair<DatumID, smart_ptr<literal::Literal> > makeLexemeDatum (smart_ptr<literal::Literal> l,
            DatumID id);
        std::pair<DatumID, smart_ptr<literal::Literal> > makeAbbreviation (smart_ptr<literal::Literal> l);

        // 渡したtreeの内容を文字列に変換する。
//        smart_ptr<unicode::UniString> treeToValue(Tree& t);
    };

};

#endif /* _TREE_H_ */
