#include <vector>
#include <string>
#include <functional>
#include <algorithm>

#include "unicode.h"

using namespace utakata;
using namespace utakata::unicode;

//================================================================================

unicode::UniChar::UniChar() : code_(0)
{
}

unicode::UniChar::UniChar(const unicode::UniChar& ch)
{
    // 単純にそれぞれの値をコピーする。
    code_ = ch.code_;
}

unicode::UniChar::UniChar(unicode::unicode_t ch) :
    code_(ch)
{
    // 安全なスワップを利用する。

}

unicode::UniChar& unicode::UniChar::operator=(const unicode::UniChar& ch)
{
    // 安全なスワップを利用する。
    UniChar c(ch);
    swap(c);
    return *this;
}

unicode::UniChar& unicode::UniChar::operator=(unicode::unicode_t ch)
{
    code_ = ch;
    return *this;
}

bool unicode::UniChar::operator==(const unicode::UniChar& ch) const
{
    return ch.code_ == code_;
}

bool unicode::UniChar::operator!=(const unicode::UniChar& ch) const
{
    return !(*this == ch);
}

bool UniChar::operator<(const UniChar& ch) const
{
    return code_ < ch.code_;
}

bool UniChar::operator>(const UniChar& ch) const
{
    return code_ > ch.code_;
}

bool unicode::UniChar::operator==(unicode::unicode_t ch) const
{
    return ch == code_;
}

bool unicode::UniChar::operator!=(unicode::unicode_t ch) const
{
    return !(*this == ch);
}

bool UniChar::operator<(unicode_t ch) const
{
    return code_ < ch;
}

bool UniChar::operator>(unicode_t ch) const
{
    return code_ > ch;
}

void unicode::UniChar::swap(unicode::UniChar& ch)
{
    // 値と標準コンテナなので、特に問題なく動作する。
    std::swap(code_, ch.code_);
}

bool unicode::is_ascii(const UniChar& ch)
{
    // 0x7f >= ascii >= 0x00 がasciiなので、その範囲で判定を行う。
    return ch.getRawCode() >= 0 && ch.getRawCode() <= 0x7f ? true : false;
}

bool unicode::is_numeric(const UniChar& ch)
{
    return '0' <= ch.getRawCode() && ch.getRawCode() <= '9' ? true : false;
}

bool unicode::is_alpha(const UniChar& ch)
{
    // 通常のアルファベットであるかどうかを判別する。
    // 大文字、小文字は問わない。
    if (('a' <= ch.getRawCode() && ch.getRawCode() <= 'z') ||
        ('A' <= ch.getRawCode() && ch.getRawCode() <= 'Z'))
    {
        return true;
    }
    return false;
}

//================================================================================

unicode::UniString::UniString() : chars_()
{
}

unicode::UniString::UniString(const UniChar& ch, int num) : chars_()
{
    chars_.assign(num, ch);
}

unicode::UniString::UniString(
    const UniString& str) : chars_(str.chars_)
{
}

void UniString::assign(const unicode::UniChar& ch, int num)
{
    UniString tmp(ch, num);
    swap(tmp);
}

void UniString::assign(const UniString& str)
{
    //基本的に=で渡した場合と全く同じなので、そのようにする。
    UniString tmp(str);
    swap(tmp);
}

UniString& UniString::operator=(const UniString& str)
{
    assign(str);
    return *this;
}

void UniString::swap(UniString& str)
{
    // シンプルにswapを行う。
    std::swap(chars_, str.chars_);
}

UniString& UniString::operator+=(const UniString& str)
{
    // 一度コピーと加算してから実際にswapさせる。
    UniString tmp(str);
    chars_.insert(chars_.end(), tmp.chars_.begin(), tmp.chars_.end());
    return *this;
}

UniString& UniString::operator+=(const UniChar& ch)
{
    // 実際には文字を設定するのに利用される。
    // また、これを定義しておくことで、streamからの結果を直接設定することができる。
    UniString tmp(*this);
    tmp.push_back(ch);
    swap(tmp);
    return *this;
}

void UniString::insert(UniString::iterator it, UniString::iterator begin,
                       UniString::iterator last)
{
    // 渡されたのは実際にはvectorのイテレータなので、
    // そのままvectorの実装に任せることができる。
    chars_.insert(it, begin, last);
}

void UniString::insert(UniString::iterator it,
                       UniString::const_iterator begin,
                       UniString::const_iterator last)
{
    // 渡されたのがconstであるかどうかというだけの違いであるため、
    // そのまま渡すことができる。
    chars_.insert(it, begin, last);
}

void UniString::push_back(const UniChar& ch)
{
    // 単純に内部の配列のpush_backを呼びだすのみで完了する。
    chars_.push_back(ch);
}

bool UniString::operator==(const UniString& str) const
{
    return chars_ == str.chars_;
}

//================================================================================

unicode::UniString unicode::substring(
    const unicode::UniString& str, size_t begin, size_t end)
{
    if (end == 0)
    {
        // 先頭から末尾までを取得する。
        UniString s;
        s.insert(s.begin(), str.begin() + begin, str.end());
        return s;
    }
    else if (begin <= end)
    {
        UniString s;
        s.insert(s.begin(), str.begin() + begin, str.begin() + (begin + end));
        return s;
    }
    else
    {
        throw range_error("out of range in unicode::substring");
    }
}

UniString unicode::operator+(const UniString& lh, const UniString& rh)
{
    // 双方をコピーして加算して返す。凄い負荷が高い。
    UniString str(lh);
    str += rh;
    return str;
}

UniString unicode::operator+(const UniString& lh, const UniChar& rh)
{
    // 文字と加算する。
    UniString tmp(lh);
    tmp += rh;
    return tmp;
}

UniString unicode::operator+(const UniChar& lh, const UniString& rh)
{
    UniString tmp(lh, 1);
    tmp += rh;
    return tmp;
}

UniString unicode::convert(const std::vector<unsigned long>& c)
{
    // 渡されたunsigned longの配列をUniCharの配列であるUniStringに
    // 変換する。
    UniString s;
    std::vector<unsigned long>::const_iterator it = c.begin(), end = c.end();
    for (; it != end; ++it)
    {
        s += UniChar(*it);
    }
    return s;
}

UniString unicode::convert(const std::string& c)
{
    // 渡された文字列がasciiであることを前提として、データの設定を行なう。
    UniString s;
    std::string::const_iterator it = c.begin(), end = c.end();
    for (; it != end; ++it)
    {
        s += UniChar(*it);
    }
    return s;
}
