#ifndef _UNICODE_H_
#define _UNICODE_H_

#include <vector>
#include <functional>
#include <string>
#include <exception>

namespace utakata {

// CUniCharの作成と、CUniCharのコンテナであるCUniStringを定義する。

    namespace unicode {

        // 内部で利用されるunicodeスカラー値の型。
        typedef unsigned long unicode_t;

        class range_error : public std::exception
        {
        public:
            range_error(const std::string& str) : str_(str) {}
            virtual ~range_error() throw() {}

            const char* what() throw() {
                return str_.c_str();
            }
        private:

            std::string str_;
        };
    
        class UniChar
        {
            // schemeにおけるunicodeスカラー値一つ分を表す。
            // スカラー値はunsigned longで保持され、基本的には構造体のように
            // 振る舞う。
        public:

            // デフォルトではnul(0x0000)が指定される。
            UniChar();
            virtual ~UniChar(){}

            // コピーコンストラクタを宣言したため、同時にoperator=を宣言する。
            UniChar(const UniChar& ch);
            UniChar(unicode_t code);
            UniChar& operator=(const UniChar& ch);
            UniChar& operator=(unicode_t ch);

            // 各演算子のオーバーロード。
            // <=と>=は、それぞれ>の結果のnot、<の結果のnotを
            // 用いることで実装できるため、この二つだけ供えていれば問題ない。
            bool operator==(const UniChar& code) const;
            bool operator<(const UniChar& code) const;
            bool operator>(const UniChar& code) const;
            bool operator!=(const UniChar& code) const;

            bool operator!=(unicode_t code) const;
            bool operator==(unicode_t code) const;
            bool operator<(unicode_t code) const;
            bool operator>(unicode_t code) const;

            // 内部で保持しているコードを返す。
            unicode_t getRawCode() const {return this->code_;}

        private:

            // 渡されたUniCharの内部と交換する。
            void swap(UniChar& ch);

            unicode_t code_;
        };

        // 各unicode用変換用の関数群。

        // 渡されたUniCharがasciiコードの範囲内に収まっているかどうかを返す。
        bool is_ascii(const UniChar& ch);

        // 渡されたUniCharが、それぞれ数値、アルファベットであるかどうか
        // を返す。これらはutf8モジュールのユーティリティ関数によって実装される。
        bool is_numeric(const UniChar& ch);
        bool is_alpha(const UniChar& ch);

        //================================================================================

        class UniString
        {
            // UniCharを複数連結して保持するためのコンテナ
            // 相互での比較などもサポートし、utakataの内部全般で
            // 標準として使用することが可能なものとし、かつ標準コンテナと
            // ほぼ同様の処理を行うことができるように定義されている。
            // というか内部では単純なvectorで定義されているため、
            // typedefされた型を指定するだけで、イテレータとして利用可能。

        public:

            // イテレータとして簡単に利用するためのtypedefマクロ
            typedef std::vector<UniChar>::iterator iterator;
            typedef std::vector<UniChar>::const_iterator const_iterator;

        public:

            UniString();
            // 第一引数のchをnum個数分設定する。
            UniString(const UniChar& ch, int num);
            UniString(const UniString& str);
            virtual ~UniString(){}

            // 実体に代入する。代入が行われなかった場合、元のデータ
            // は保存される。
            void assign(const UniChar& ch, int num);
            void assign(const UniString& str);

            // iteratorを取得する。
            const_iterator begin() const {return chars_.begin();}
            iterator begin() {return chars_.begin();}

            // 末尾のイテレータを取得する。
            const_iterator end() const {return chars_.end();}
            iterator end() {return chars_.end();}

            // []をオーバーロードする。これは常に境界のチェックを行う。
            const UniChar& operator[](size_t t) const {return chars_.at(t);}
            UniChar& operator[](size_t t) {
                return const_cast<UniChar&>(
                    static_cast<const UniString*>(this)->chars_[t]);}

            // サイズを取得する。
            size_t size() const {return chars_.size();}
            // 空かどうかを調べる。
            bool empty() const {return chars_.empty();}

            // 安全なswapと共に提供する。
            UniString& operator=(const UniString& str);

            // 加算のみをサポートする。
            UniString& operator+=(const UniString& str);
            UniString& operator+=(const UniChar& ch);
            UniString& operator+=(const std::string& str);

            bool operator==(const UniString& str) const;
            bool operator!=(const UniString& str) const {
                return !(*this == str);
            }

            // 渡したイテレータの範囲を追加する
            // beginとlastは同一のUniStringから取得されたものであること。
            void insert(iterator it, iterator begin, iterator last);
            void insert(iterator it, const_iterator begin,
                        const_iterator last);

            // UniCharを単体で末尾に追加する。
            void push_back(const UniChar& ch);

        private:

            void swap(UniString& str);

            std::vector<UniChar> chars_;
        };

        UniString operator+(const UniString& lh, const UniString& rh);
        UniString operator+(const UniString& lh, const UniChar& rh);
        UniString operator+(const UniChar& lh, const UniString& rh);

        // substringの実装を行う。
        // [begin, end)までの文字を文字列として返す。
        // endが渡されないか、0が渡された場合、beginから末尾までが返される。
        UniString substring(const UniString& str, size_t begin, size_t end = 0);

        // 渡したunsigned longの配列をUniCharの配列であるUniStringに変換する。
        UniString convert(const std::vector<unsigned long>& d);
        // 渡したstringが、全てascii文字列であることを前提として、データの
        // 設定を行う。
        UniString convert(const std::string& d);


    };

};

#endif /* _UNICODE_H_ */
