#include <algorithm>
#include <sstream>
#include <assert.h>

#include "utf8_transcoder.h"
#include "smart_ptr.h"

using namespace std;
using namespace utakata::transcoder;
using namespace utakata;

transcoder::UTF8Transcoder::UTF8Transcoder() : utf8_max_bit_pos(21)
{
}

unsigned long transcoder::UTF8Transcoder::encode(std::istream& p)
{
    // 渡されたバイナリポートから、一文字を読みとる。
    // 最初に一文字だけ読みだして、チェックをかける。

    // まずは読み取りをおこなう。
    int c = 0;
    c = p.peek();

    std::vector<int> bytes;
    if (c != std::istream::traits_type::eof())
    {
        // 末尾でない場合のみ、以降のチェックに入る。
        size_t size = 0;
        // 先頭1バイトが正常でなかった場合はそのまま抜ける
        if (utf8::is_utf8_first_byte(c, size))
        {
            bytes.push_back(p.get());
            if ((size - 1) > 0)
            {
                // sizeが0より大きい場合には、この時複数バイトで文字が構成
                // されていると考えられるため、明示的に複数文字を取得する。
                // peekされているので、読み取り位置は変化しない。
                for (size_t i = 0; i < size - 1; ++i)
                {
                    bytes.push_back(p.get());
                }
            }
        }
    }

    return generate(bytes);
}

unsigned long transcoder::UTF8Transcoder::generate(const std::vector<int>& bytes)
{
    // 1文字分のUTF8のバイト列を受け取って、コードに変換して返す。
    // 先頭の値によって、次のように値を決定することができる。
    // x = utf8の先頭バイト
    // y1〜yN = utf8の先頭バイト以降のバイト
    // N = utf8の先頭バイトを含むバイト数
    // code = (y1 & ((1 << 7) - 1)) << (6 * n-1) + (y2 & ((1 << 7) -1)) << (6 * (n - 1))...+ x & ((1 << N) -1) << (6 * N-1)
    // 先頭バイト以外は、全て先頭に10とうビットが設定されている。このビットを除いた6ビットをする。
    // つまり、末尾のバイトから順次やっていけばよい。

    const unsigned char max_c = (1 << (sizeof(unsigned char) * 8 - 1)) - 1;
    // 固定の範囲外である値を設定しておく。
    unsigned long code = 0x110000;

    // 一文字も読みだせていないなら0を返す。
    if (!bytes.empty())
    {
        code = 0;
        if (bytes.size() == 1)
        {
            // asciiコードは7bitなのでそこだけ切り取って返す。
            code = bytes[0] & max_c;
        }
        else
        {
            // 一時的に利用されるクラス。
            struct Lambda
            {
                unsigned long operator()(unsigned char c, int s) {
                    return c << (6 * s);
                }
            };

            std::vector<int> cl = bytes;
            int first = cl[0];
            cl.erase(cl.begin());
            // サイズが1以外の場合、ここからがちと違う。
            std::vector<int>::const_reverse_iterator beg = cl.rbegin(),
                end = cl.rend();
            const unsigned char char_bit = (1 << 6) - 1;
            for (int i = 0; beg != end; ++i,++beg)
            {
                code += Lambda()((*beg & char_bit), i);
            }

            // 最後だけ、別の計算が必要になる。
            const unsigned char byte_size = sizeof(unsigned char) * 8;
            const unsigned char first_byte = (1 << (byte_size - bytes.size() - 1)) - 1;
            code += Lambda()(first & first_byte, bytes.size() - 1);
        }
    }

    return code;
}

std::vector<int> utakata::transcoder::UTF8Transcoder::decode(unsigned long ch)
{
    // encodeされたunicode文字列を、UTF8に変更する。
    // 基本的には単純な逆変換のようなものになるため、それほど難しくは無いと
    // 思いたい。

    // まずは抽出したデータの最上位bitを抽出する。
    // utf-8のため、最大21bitなので、20bitシフトした時点から始める。

    unsigned long max_bit = 1 << (utf8_max_bit_pos - 1);
    int pos = utf8_max_bit_pos;
    while (max_bit > 0 && !(ch & max_bit))
    {
        // 先頭bitが1ではない場合には、1bit右へシフトする。
        max_bit >>= 1;
        --pos;
    }

    if (max_bit == 0)
    {
        // この場合、null文字であるはずなのでこれを出力する。
        return std::vector<int>(1, 0x0);
    }

    // 先頭bitが無事発見できた場合には、bit位置と現在位置の兼ね合いから、
    // それぞれbitを設定していく。
    std::vector<int> ret;
    const int utf8_byte_base = 0x80;
    const int byte_base_bit_size = 6;
    const int utf8_byte_mask = (1 << byte_base_bit_size) - 1;
    unsigned long target = ch;
    for (int i = pos; i > byte_base_bit_size; i -= byte_base_bit_size)
    {
        // それぞれ算出していく。
        ret.push_back(utf8_byte_base | (target & utf8_byte_mask));
        target >>= byte_base_bit_size;
    }

    // 残ったbit分を、それまでに算出した文字の分+1だけ1を先頭に付けてやる。
    {
        const int bit_first_head = ((1 << (ret.size() + 2)) - 1) - 1;
        const int byte_size = sizeof(unsigned char) * 8;
        // 上の例は,1 << 3 = 1000 - 1 = 0111 - = 0110 みたいな感じ。

        // 先頭のbitを、retのサイズ + 2したのをbyte_sizeから引いて、
        // その値だけfirst_headをシフトさせて、それに残っている
        // targetを or する。
        // 110 << ( 8 - 1 - 2 ) = 11000000 | 11100 = 11011100
        ret.push_back((bit_first_head << (byte_size - ret.size() - 2)) |
                      target);
     }

    // 結果はそのまま反転させることで、正しい形となる。
    std::reverse(ret.begin(), ret.end());
    return ret;
}

// utf8関連の便利関数

bool utakata::transcoder::utf8::is_utf8_one(const std::vector<int>& bytes, size_t& size)
{
    //渡したバイト列がUTF8の一文字に該当するかどうかを返す。
    size_t back = size;
    
    if (bytes.size() == 0)
    {
        size = back;
        return false;
    }

    size_t num = 0;

    if (!is_utf8_first_byte(bytes[0], num))
    {
        size = back;
        return false;
    }

    // そもそもbytesのサイズが足りない場合にも失敗とする。
    if (num > bytes.size())
    {
        size = back;
        return false;
    }

    // 先頭要素以外が正しければそれで問題ないとする。
    if (num > 1)
    {
        
        const CheckUTF8Byte& checker = for_each(bytes.begin() + 1, bytes.begin() + num,
                                                 CheckUTF8Byte());
        
        if (checker.good)
        {
            size = num;
        } else {
            size = back;
            return false;
        }
    } else {
        // sizeが0の場合には、この時点で1を設定するようにする。
        size = num;
    }

    // 最後まで到達した時点で成功とする。
    return true;
}

bool utakata::transcoder::utf8::is_utf8_all(const std::vector<int>& bytes)
{
    // 与えられたバイト列全てがUTF-8であるかどうかを返す。
    size_t size = 0;
    std::vector<int>::const_iterator it = bytes.begin();
    while (is_utf8_one(std::vector<int>(it, bytes.end()), size)) {
        it += size;
        size = 0;
    }

    if (it == bytes.end()) {
        return true;
    }

    return false;
}

bool utakata::transcoder::utf8::is_utf8_first_byte(int c, size_t& size)
{
    // UTf-8の先頭バイトであるかどうかを返す。
    // 先頭バイトである場合には、その先頭バイトを含む、一文字のサイズを返す。
    // 先頭バイトではない場合には、
    const unsigned char max_c = 1 << (sizeof(unsigned char) * 8 - 1);

    size_t back = size;

    // 最上位ビットが0である場合、これはasciiコードを指す。
    if (!(c & max_c))
    {
        size = 1;
        return true;
    }

    unsigned char first = c << 1;
    size_t num = 1;
    while (first & max_c) {
        first <<= 1;
        num += 1;
    }

    // ここまできたとき、最上位ビットは0であるはず。
    // numが5未満である場合、とりあえず正常としておくこととする。
    const unsigned char max_utf8_sequence = 5;
    if (num < max_utf8_sequence) {
        size = num;
        return true;
    }
    else if (num == 1)
    {
        // numが1の場合、何らかの理由で先頭が欠落したと見られる。
        // この場合、スキップするべきバイト数を返す。
        size = 1;
        return false;
    
    } else {
        size = back;
        return false;
    }
}


bool utakata::transcoder::utf8::is_utf8_ascii(const std::vector<int>& bytes)
{
    // 一文字かつ、0x00〜0x7fの範囲であるデータであることが条件となる。
    size_t s = 0;
    bool b = is_utf8_one(bytes, s);

    if (b && s == 1)
    {
        return true;
    }
    return false;
}

bool utakata::transcoder::utf8::is_utf8_numeric(const std::vector<int>& bytes)
{
    // 一文字分だけが渡されていると判断する。
    if (!is_utf8_ascii(bytes))
    {
        return false;
    }

    if (bytes[0] >= '0' && bytes[0] <= '9')
    {
        return true;
    }

    return false;
}

bool utakata::transcoder::utf8::is_utf8_alpha(const std::vector<int>& bytes)
{
    // 同じく一文字であると判別する。
    // asciiのサブセットなので、先にasciiであると判別しておく。
    if (!is_utf8_ascii(bytes))
    {
        return false;
    }

    if ((bytes[0] >= 'a' && bytes[0] <= 'z') ||
        (bytes[0] >= 'A' && bytes[0] <= 'Z'))
        
    {
        return true;
    }

    return false;
}
