#ifndef _UTF8_TRANSCODER_H_
#define _UTF8_TRANSCODER_H_

#include <iostream>
#include <string>
#include <vector>

#include "smart_ptr.h"
#include "transcoder.h"

namespace utakata {

    namespace transcoder {

        class UTF8Transcoder : public ITranscoder
        {
            // UTF-8のバイト列が与えられた場合に利用されるTranscoder

            unsigned long utf8_max_bit_pos;

        public:
            UTF8Transcoder();
            virtual ~UTF8Transcoder() {}

            virtual unsigned long encode(std::istream& p);
            virtual std::vector<int> decode(unsigned long code);

        private:

            unsigned long generate(const std::vector<int>& bytes);

        };

        namespace utf8 {

            // これらの便利関数は、とりあえずこの中に定義しておく。

            struct CheckUTF8Byte
            {
                // UTF8の先頭バイト以外であるかどうかをチェックする。
                const int checker;
                bool good;
                CheckUTF8Byte() : checker(0x2), good(true) {}

                template<class T>
                void operator()(const T& t) {
                    // 先頭ビットが10ではない場合、チェックに失敗する。
                    T tmp = t >> 6;
                    if ( ((tmp & 0x3)) != checker) {
                        good = false;
                    }
                }
            };

            struct PutBack
            {
                // 渡されたデータをistreamにputbackする。
                std::istream& strm_;
                PutBack(std::istream& strm) : strm_(strm) {}

                template<class T>
                void operator()(T t)
                    {
                        strm_.unget(t);
                    }
            };
        
            // 与えられたバイト列の先頭から、UTF8一文字に該当しているかどうかを返す。
            // バイト列がUTF8に該当する場合、そのバイト列のサイズを返す。
            bool is_utf8_one(const std::vector<int>& bytes, size_t& size);

            // 与えられたバイト列全てが、UTF8に該当しているかどうかを返す。
            bool is_utf8_all(const std::vector<int>& bytes);

            // UTF-8の先頭バイトとして正しいフォーマットであるかどうか。
            // 正しいフォーマットである場合、渡したバイトを含めた、一文字である
            // バイト数を返す。
            bool is_utf8_first_byte(int c, size_t& size);

            // 渡したバイト列が、0-9で判別される数値文字と一致するかどうかを返す。
            // is_asciiのサブセットを判別するものである。
            bool is_utf8_numeric(const std::vector<int>& bytes);

            // 渡したバイト列が、a-zA-Zの領域であるかどうかを調べる。
            // is_asciiのサブセットを判別するものである。
            bool is_utf8_alpha(const std::vector<int>& bytes);

            // 渡したバイト列が、0x00〜0x7fの領域であるかどうかを返す。
            // is_utf8_alpha, is_utf8_numericのスーパーセットとなる。
            bool is_utf8_ascii(const std::vector<int>& bytes);
        };
    };
}

#endif /* _UTF8_TRANSCODER_H_ */
